import os
import sys
import subprocess
import shutil
from importlib.util import find_spec

from jinja2 import Template, Environment, FileSystemLoader
from django.conf import settings
import logging
logger = logging.getLogger(__name__)

class ScriptTemplate:
    def __init__(self, template_top, template_path):
        templ_path = os.path.join(template_top, template_path)
        template_dir = os.path.dirname(templ_path)
        template_name = os.path.basename(templ_path)
        env = Environment(loader=FileSystemLoader(template_dir))
        self._template = env.get_template(template_name)
        logger.debug(f"Loaded job template at {templ_path}")

    def render(self, qlaunch):
        conf = self.qlaunch_to_dict(qlaunch)
        return self._template.render(conf)

    def qlaunch_to_dict(self, qlaunch):
        if qlaunch.project:
            project = qlaunch.project
        else:
            project = settings.DEFAULT_PROJECT
        if qlaunch.wf_filter:
            wf_filter = f'wf-filter={qlaunch.wf_filter}'
        else:
            wf_filter = 'consume-all'
        conf = dict(project=project,
                    queue=qlaunch.queue,
                    nodes=qlaunch.nodes,
                    time_minutes=qlaunch.wall_minutes,
                    job_mode=qlaunch.job_mode,
                    wf_filter=wf_filter,
                    )
        balsam_env = self.get_balsam_env()
        conf.update(balsam_env)
        return conf

    @staticmethod
    def get_balsam_env():
        balsam_bin = shutil.which("balsam")
        pg_bin = os.path.dirname(shutil.which("pg_ctl"))
        if balsam_bin is None:
            balsam_bin = os.path.dirname(sys.executable)
        else:
            balsam_bin = os.path.dirname(balsam_bin)

        balsam_db_path = os.environ['BALSAM_DB_PATH']
        return dict(
            balsam_bin=balsam_bin,
            balsam_db_path=balsam_db_path,
            pg_bin=pg_bin,
        )



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import balsam.launcher.dag as dag

print("Hello from reduce_post")

for i in range(5):
    with open(f"summary{i}.dat", 'w') as fp:
        fp.write("test\n")



import sys
import balsam.launcher.dag as dag
import time

start = time.time()

while dag.BalsamJob.objects.filter(name="slow_job").count() == 0:
    time.sleep(2)
    if time.time() - start > 40:
        raise RuntimeError("the slow job never started")

slow_job = dag.BalsamJob.objects.get(name='slow_job')

if 'when-running' in ' '.join(sys.argv):
    while slow_job.state != 'RUNNING':
        time.sleep(2)
        slow_job.refresh_from_db()
dag.kill(slow_job)



from balsam.django_config import serverinfo
from .infolock import InfoLock
import sys
import time
import socket
import signal
import subprocess
import os

class SignalReceived(Exception): pass
def term_handler(signum, stack):
    raise SignalReceived(f"Killed by signal {signum}")

def launch_server(info):
    if not info._is_owner:
        raise PermissionError("Please ask the owner of this DB to launch it")
    info.reset_server_address()
    log_path = os.path.join(info.balsam_db_path, 'postgres.log')
    user_home = os.path.join(os.path.expanduser('~'), '.balsam')
    if not os.path.exists(user_home):
        print("Creating ~/.balsam directory")
        os.makedirs(user_home)
    start_cmd = f"pg_ctl -w start -D {info.pg_db_path} -l {log_path} --mode=smart"
    print("Launching Balsam DB server")
    proc = subprocess.run(start_cmd, shell=True, check=True)

def kill_server(info):
    if not info._is_owner:
        raise PermissionError("Only the owner of the server-info can kill this Balsam DB")
    local_host = socket.gethostname()
    server_host = info['host']
    stop_cmd = f"pg_ctl -w stop -D {info.pg_db_path} --mode=smart"

    if server_host == local_host:
        print("Stopping local Balsam DB server")
        subprocess.run(stop_cmd, shell=True)
        time.sleep(1)
    else:
        print("Stopping remote Balsam DB server")
        subprocess.run(f"ssh {server_host} {stop_cmd}", shell=True)
    info.update(dict(host='', port=''))


def test_connection(info, raises=False):
    from balsam.launcher import dag
    from django.conf import settings
    from django import db

    # reset connection in case serverinfo changed *after* Django setup
    db.connections.close_all()
    settings.DATABASES['default']['PORT'] = info['port']
    settings.DATABASES['default']['HOST'] = info['host']

    try:
        count = dag.BalsamJob.objects.count()
    except db.OperationalError:
        if raises: raise
        else: return False
    else:
        assert type(count) is int
        return True

def reset_main(db_path):
    start_main(db_path)
    info = serverinfo.ServerInfo(db_path)
    kill_server(info)

def start_main(db_path):
    old_handler = signal.getsignal(signal.SIGTERM)
    signal.signal(signal.SIGTERM, term_handler)
    with InfoLock(db_path):
        _start_main(db_path)
    signal.signal(signal.SIGTERM, old_handler)

def _start_main(db_path):
    info = serverinfo.ServerInfo(db_path)

    if not (info.get('host') or info.get('port')):
        launch_server(info)
        test_connection(info, raises=True)
    elif test_connection(info):
        print("Connected to already running Balsam DB server!")
    elif info._is_owner:
        print(f"Server at {info['host']}:{info['port']} isn't responsive; will try to kill and restart")
        kill_server(info)
        launch_server(info)
        test_connection(info, raises=True)
    else:
        print(f"Server at {info['host']}:{info['port']} isn't responsive; please ask the owner to restart it")

def list_connections(db_path):
    info = serverinfo.ServerInfo(db_path)
    host = info['host']
    port = info['port']
    subprocess.run(
        f'''psql -d balsam -h {host} -p {port} -c \
        "SELECT pid,application_name,usename,state,substr(query, 1, 60)\
        FROM pg_stat_activity WHERE datname = 'balsam';" \
        ''',
        shell=True
    )

def add_user(db_path, uname):
    info = serverinfo.ServerInfo(db_path)
    host = info['host']
    port = info['port']
    subprocess.run(
        f'''psql -d balsam -h {host} -p {port} -c \
        "CREATE user {uname}; \
        grant all privileges on database balsam to {uname}; \
        GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA public TO {uname}; \
        GRANT ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA public TO {uname}; "\
        ''',
        shell=True
    )

def drop_user(db_path, uname):
    info = serverinfo.ServerInfo(db_path)
    host = info['host']
    port = info['port']
    subprocess.run(
        f'psql -d balsam -h {host} -p {port} -c \
        "REVOKE ALL privileges on all tables in schema public FROM {uname};\
        REVOKE ALL privileges on all sequences in schema public FROM {uname};\
        REVOKE ALL privileges on database balsam FROM {uname};\
        DROP ROLE {uname};"',
        shell=True
    )

def list_users(db_path):
    info = serverinfo.ServerInfo(db_path)
    host = info['host']
    port = info['port']
    subprocess.run(
        f'''psql -d balsam -h {host} -p {port} -c \
        "SELECT rolname FROM pg_roles;"
        ''',
        shell=True
    )

def create_db(serverInfo):
    db_path = serverInfo.pg_db_path
    retcode = subprocess.Popen(f'initdb -D {db_path} -U $USER', shell=True).wait()
    if retcode != 0: raise RuntimeError("initdb process failed")
    with open(os.path.join(db_path, 'postgresql.conf'), 'a') as fp:
        fp.write("listen_addresses = '*' # appended from balsam init\n")
        fp.write('port=0 # appended from balsam init\n')
        fp.write('max_connections=128 # appended from balsam init\n')
        fp.write('shared_buffers=2GB # appended from balsam init\n')
        fp.write('synchronous_commit=off # appended from balsam init\n')
        fp.write('wal_writer_delay=400ms # appended from balsam init\n')
        fp.write('logging_collector=on # appended from balsam init\n')
        fp.write('log_min_duration_statement=0 # appended from balsam init\n')
        fp.write('log_connections=on # appended from balsam init\n')
        fp.write('log_duration=on # appended from balsam init\n')
    with open(os.path.join(db_path, 'pg_hba.conf'), 'a') as fp:
        fp.write(f"host all all 0.0.0.0/0 trust\n")
    launch_server(serverInfo)
    serverInfo.refresh()
    port = serverInfo['port']
    create_proc = subprocess.Popen(f'createdb balsam -p {port}', shell=True)
    retcode = create_proc.wait()
    if retcode != 0: raise RuntimeError("createdb failed")
    else: print("Created `balsam` DB")



from subprocess import Popen, PIPE, STDOUT
import os
from threading import Thread
from django.conf import settings
import time

def time_cmd(cmd, stdout=PIPE, stderr=STDOUT, envs=None):
    '''Return string output from a command line'''
    if type(cmd) == list:
        cmd = ' '.join(cmd)

    cmd = f'time ( {cmd} )'
    p = subprocess.Popen(cmd, shell=True, stdout=stdout,
                         stderr=stdout, env=envs,
                         executable='/bin/bash')
    stdout = p.communicate()[0].decode('utf-8')
    real_seconds = parse_real_time(stdout)
    return stdout, realtime

def parse_real_time(stdout):
    '''Parse linux "time -p" command'''
    if type(stdout) == bytes:
        stdout = stdout.decode()

    lines = stdout.split('\n')

    real_lines = [l for l in lines[-5:] if l.startswith('real')]
    if not real_lines:
        return None
    elif len(real_lines) > 1:
        real_line = real_lines[-1]
    else:
        real_line = real_lines[0]

    time_str = real_line.split()[1]
    return float(time_str)

def get_tail(fname, nlines=16, indent='    '):
    '''grab last nlines of fname and format nicely'''

    # Easier to ask OS than implement a proper tail
    proc = Popen(f'tail -n {nlines} {fname}'.split(),stdout=PIPE, 
                 stderr=STDOUT)
    tail = proc.communicate()[0].decode()
    lines = tail.split('\n')
    for i, line in enumerate(lines[:]):
        lines[i] = indent + line
    return '\n'.join(lines)

class cd:
    '''Context manager for changing cwd'''
    def __init__(self, new_path):
        self.new_path = os.path.expanduser(new_path)

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.new_path)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            print(exc_type, exc_value, traceback)
        os.chdir(self.saved_path)

class MonitorStream(Thread):
    '''Thread: non-blocking read of a process's stdout'''
    def __init__(self, runner_output):
        super().__init__()
        self.stream = runner_output
        self.queue = Queue()
        self.daemon = True

    def run(self):
        # Call readline until empty string is returned
        for line in iter(self.stream.readline, b''):
            self.queue.put(line.decode('utf-8'))
        self.stream.close()

    def available_lines(self):
        while True:
            try: yield self.queue.get_nowait()
            except Empty: return

def delay_generator(period=settings.SERVICE_PERIOD):
    '''Generator: Block for ``period`` seconds since the last call to __next__()'''
    nexttime = time.time() + period
    while True:
        now = time.time()
        tosleep = nexttime - now
        if tosleep <= 0:
            nexttime = now + period
        else:
            time.sleep(tosleep)
            nexttime = now + tosleep + period
        yield

def elapsed_time_minutes():
    '''Generator: yield elapsed time in minutes since first call to __next__'''
    start = time.time()
    while True:
        yield (time.time() - start) / 60.0

def remaining_time_minutes(time_limit_minutes=0.0):
    '''Generator: yield remaining time for Launcher execution

    If time_limit_minutes is given, use internal timer to count down remaining
    time. Otherwise, query scheduler for remaining time.

    Args:
        - ``time_limit_minutes`` (*float*): runtime limit
    Yields:
        - ``remaining`` (*float*): remaining time
    Raises:
        - ``StopIteration``: when there is no time left
    '''
    if time_limit_minutes > 0.0:
        elapsed_timer = elapsed_time_minutes()
        get_remaining = lambda: time_limit_minutes - next(elapsed_timer)
    else:
        from balsam.service.schedulers import JobEnv
        get_remaining = lambda: JobEnv.remaining_time_seconds() / 60.0

    while True:
        remaining_min = get_remaining()
        if remaining_min > 0: yield remaining_min
        else: raise StopIteration



'''mpi4py wrapper that allows an ensemble of serial applications to run in
parallel across ranks on the computing resource'''
import argparse
from collections import defaultdict
import os
import sys
import logging
import random
from subprocess import Popen, STDOUT, TimeoutExpired
import shlex
import signal
import time

from mpi4py import MPI
from django.db import transaction, connections

from balsam import config_logging, settings, setup
setup()
from balsam.launcher.exceptions import *
from balsam.launcher.util import cd, get_tail, remaining_time_minutes
from balsam.core.models import BalsamJob, safe_select

logger = logging.getLogger('balsam.launcher.mpi_ensemble')
config_logging('serial-launcher')

comm = MPI.COMM_WORLD
RANK = comm.Get_rank()
MSG_BUFSIZE = 2**16
connections.close_all()

class ResourceManager:

    FETCH_PERIOD = 2.0
    KILLED_REFRESH_PERIOD = 3.0

    def __init__(self, job_source):
        self.job_source = job_source
        self.node_occupancy = [0.0 for i in range(comm.size)]
        self.node_occupancy[0] = 1.0
        self.running_locations = {}
        self.job_occupancy = {}

        self.last_job_fetch = -10.0
        self.last_killed_refresh = -10.0
        self.job_cache = []
        self.killed_pks = []

        self.recv_requests = {i:comm.irecv(MSG_BUFSIZE, source=i) for i in range(1,comm.size)}

        self.job_source.check_qLaunch()
        if self.job_source.qLaunch is not None:
            sched_id = self.job_source.qLaunch.scheduler_id
            self.RUN_MESSAGE = f'Scheduled by Balsam Service (Scheduler ID: {sched_id})'
        else:
            self.RUN_MESSAGE = 'Not scheduled by Balsam service'
        logger.info(self.RUN_MESSAGE)
        logger.info(f'Assigning jobs to {comm.size-1} worker ranks')

    def refresh_job_cache(self):
        now = time.time()
        if len(self.job_cache) == 0 or (now-self.last_job_fetch) > self.FETCH_PERIOD:
            jobquery = self.job_source.get_runnable(
                max_nodes=1,
                serial_only=True,
                order_by=('node_packing_count', # ascending
                          '-wall_time_minutes') # descending
            )
            self.job_cache = list(jobquery[:10000])
            self.last_job_fetch = now
            logger.debug(f"Refreshed job cache: {len(self.job_cache)} runnable")
            if len(self.job_cache) == 0:
                logger.debug(f'Job cache query\n{jobquery.query}\n')

    def refresh_killed_jobs(self):
        now = time.time()
        if now - self.last_killed_refresh > self.KILLED_REFRESH_PERIOD:
            killed_pks = self.job_source.filter(state='USER_KILLED').values_list('job_id', flat=True)

            if len(killed_pks) > len(self.killed_pks): 
                logger.info(f"Killed jobs: {self.killed_pks}")
            self.killed_pks = killed_pks
            self.last_killed_refresh = now
        
    def pre_assign(self, rank, job):
        job_occ = 1.0 / job.node_packing_count
        self.node_occupancy[rank] += job_occ
        self.job_occupancy[job.pk] = job_occ
        self.running_locations[job.pk] = rank

    def revert_assign(self, rank, job_pk):
        job_occ = self.job_occupancy[job_pk]
        self.node_occupancy[rank] -= job_occ
        if self.node_occupancy[rank] < 0.0001: self.node_occupancy[rank] = 0.0
        del self.job_occupancy[job_pk]
        del self.running_locations[job_pk]

    @transaction.atomic
    def allocate_next_jobs(self):
        '''Generator: yield (job,rank) pairs and mark the nodes/ranks as busy'''
        self.refresh_job_cache()
        send_requests = []
        pre_assignments = defaultdict(list)
        min_packing_count = 1

        for job in self.job_cache:
            if job.node_packing_count < min_packing_count: continue
            job_occ = 1.0 / job.node_packing_count
            
            free_ranks = (i for i in range(1, comm.size) 
                          if self.node_occupancy[i]+job_occ < 1.0001)
            rank = next(free_ranks, None)

            if rank is None:
                logger.debug(f'no free ranks to assign {job.cute_id}')
                min_packing_count = job.node_packing_count + 1
            else:
                pre_assignments[rank].append(job)
                self.pre_assign(rank, job)

        if len(pre_assignments) == 0: return False

        to_acquire = [job.pk for rank in pre_assignments 
                      for job in pre_assignments[rank]]
        acquired_pks = self.job_source.acquire(to_acquire)
        logger.info(f'Acquired lock on {len(acquired_pks)} out of {len(to_acquire)} jobs marked for running')

        # Make actual assignment:
        for (rank, pre_jobs) in pre_assignments.items():
            runjobs = []
            for j in pre_jobs:
                if j.pk in acquired_pks: 
                    runjobs.append(j)
                    self.job_cache.remove(j)
                else:
                    self.revert_assign(rank, j.pk)

            if runjobs:
                mpiReq = self._send_jobs(runjobs, rank)
                logger.info(f"Sent {len(runjobs)} jobs to rank {rank}: occupancy is now {self.node_occupancy[rank]}")
                send_requests.append(mpiReq)

        BalsamJob.batch_update_state(acquired_pks, 'RUNNING', self.RUN_MESSAGE)
        logger.debug("allocate_next_jobs: waiting on all isends...")
        MPI.Request.waitall(send_requests)
        logger.debug("allocate_next_jobs: all isends completed.")
        return len(acquired_pks) > 0

    def _send_jobs(self, jobs, rank):
        '''Send message to compute rank'''
        message = {}
        message['tag'] = 'NEW'
        for job in jobs:
            job_spec = dict(
                workdir=job.working_directory,
                name=job.name,
                cuteid=job.cute_id,
                cmd=job.app_cmd,
                envs=job.get_envs(),
                envscript=job.envscript,
            )
            message[job.pk] = job_spec

        req = comm.isend(message, dest=rank)
        return req

    def _get_requests(self):
        completed_requests = []
        stat = MPI.Status()
        for rank in self.recv_requests:
            req = self.recv_requests[rank]
            logger.debug(f"calling req.test() on rank {rank}'s request...")
            done, msg = req.test(status = stat)
            logger.debug(f"req.test() call completed:\ndone = {done}\nmsg = {msg}")
            if done: 
                completed_requests.append((stat.source, msg))
                assert stat.source == rank

        for rank,msg in completed_requests:
            self.recv_requests[rank] = comm.irecv(MSG_BUFSIZE, source=rank)
        return completed_requests

    def serve_requests(self):
        requests = self._get_requests()
        done_jobs = []
        error_jobs = []
        killed_pks = []
        send_reqs = []
        for rank, msg in requests:
            kill_pks, req = self._handle_ask(rank, msg['ask'])
            killed_pks.extend(kill_pks)
            send_reqs.append(req)
            done_jobs.extend(msg['done'])
            error_jobs.extend(msg['error'])

        if done_jobs:  self._handle_dones(done_jobs)
        if error_jobs: self._handle_errors(error_jobs)
        if killed_pks: self.job_source.release(killed_pks)
        logger.debug("serve_requests: waiting on all isends...")
        MPI.Request.waitall(send_reqs)
        logger.debug("serve_requests: all isends completed.")
        return len(requests)
        
    def _handle_ask(self, rank, ask_pks):
        self.refresh_killed_jobs()
        response = {'tag': 'CONTINUE', 'kill_pks': []}

        for pk in ask_pks:
            if pk in self.killed_pks:
                response['tag'] = 'KILL'
                response['kill_pks'].append(pk)
        req = comm.isend(response, dest=rank)

        for pk in response['kill_pks']:
            self.revert_assign(rank, pk)

        if response['tag'] == 'KILL':
            logger.info(f"Sent KILL to rank {rank} for {response['kill_pks']}\n"
                         f"occupancy is now {self.node_occupancy[rank]}")

        return response['kill_pks'], req
    
    def _handle_dones(self, done_pks):
        for pk in done_pks:
            rank = self.running_locations[pk]
            self.revert_assign(rank, pk)

        BalsamJob.batch_update_state(done_pks, 'RUN_DONE')
        self.job_source.release(done_pks)
        logger.info(f"RUN_DONE: {len(done_pks)} jobs")
    
    @transaction.atomic
    def _handle_errors(self, error_jobs):
        error_pks = [j[0] for j in error_jobs]
        safe_select(BalsamJob.objects.filter(pk__in=error_pks))
        for pk,retcode,tail in error_jobs:
            rank = self.running_locations[pk]
            self.revert_assign(rank, pk)
            job = BalsamJob.objects.get(pk=pk)
            state_msg = f"nonzero return {retcode}: {tail}"
            job.update_state('RUN_ERROR', state_msg)
        self.job_source.release(error_pks)
    
    def send_exit(self):
        logger.debug(f"send_exit: waiting on all pending recvs")
        active_ranks = list(set(self.running_locations.values()))
        requests = [self.recv_requests[i] for i in active_ranks]
        MPI.Request.waitall(requests)
        reqs = []
        logger.debug(f"send_exit: send EXIT tag to all ranks")
        for i in range(1, comm.size):
            req = comm.isend({'tag': 'EXIT'}, dest=i)
            reqs.append(req)
        MPI.Request.waitall(reqs)

class Master:
    def __init__(self):
        self.MAX_IDLE_TIME = 20.0
        self.DELAY_PERIOD = 1.0
        self.idle_time = 0.0
        self.EXIT_FLAG = False

        args = self.parse_args()
        comm.bcast(args.gpus_per_node, root=0)
        self.remaining_timer = remaining_time_minutes(args.time_limit_min)
        next(self.remaining_timer)
        
        job_source = BalsamJob.source
        job_source.workflow = args.wf_name
        job_source.start_tick()
        job_source.clear_stale_locks()
        self.manager = ResourceManager(job_source)

        if job_source.workflow:
            logger.info(f'MPI Ensemble pulling jobs with WF {args.wf_name}')
        else:
            logger.info('MPI Ensemble consuming jobs matching any WF name')

    def parse_args(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--wf-name')
        parser.add_argument('--time-limit-min', type=float, default=72.*60)
        parser.add_argument('--gpus-per-node', type=int, default=0)
        return parser.parse_args()

    def exit(self):
        outstanding_job_pks = list(self.manager.running_locations.keys())
        num_timeout = len(outstanding_job_pks)
        logger.info(f"Shutting down with {num_timeout} jobs still running..timing out")
        BalsamJob.batch_update_state(outstanding_job_pks, 'RUN_TIMEOUT', 'timed out in MPI Ensemble')
        self.manager.job_source.release_all_owned()
        self.manager.send_exit()
        logger.debug("Send_exit: master done")
        logger.info(f"master calling MPI Finalize")
        MPI.Finalize()
        logger.info(f"ensemble master exit gracefully")
        sys.exit(0)

    def main(self):
        for remaining_minutes in self.remaining_timer:
            logger.debug(f"{remaining_minutes} minutes remaining")
            self._main()
            if self.EXIT_FLAG:
                logger.info("EXIT_FLAG on; master breaking main loop")
                break
            if self.idle_time > self.MAX_IDLE_TIME and not self.manager.running_locations:
                logger.info(f"Nothing to do for {self.MAX_IDLE_TIME} seconds: quitting")
                break
        self.exit()

    def _main(self):
        ran_anything = False
        got_requests = 0

        ran_anything = self.manager.allocate_next_jobs()
        start = time.time()
        got_requests = self.manager.serve_requests()
        elapsed = time.time() - start
        if got_requests: logger.debug(f"Served {got_requests} requests in {elapsed:.3f} seconds")

        if not (ran_anything or got_requests):
            time.sleep(self.DELAY_PERIOD)
            self.idle_time += self.DELAY_PERIOD
        else:
            self.idle_time = 0.0

class FailedToStartProcess:
    returncode = 12345
    def wait(self, timeout=0): return 12345
    def poll(self, timeout=0): return 12345
    def communicate(self, timeout=0): pass
    def terminate(self): pass
    def kill(self): pass

class Worker:
    CHECK_PERIOD=10
    RETRY_WINDOW = 20
    RETRY_CODES = [-11, 1, 255, 12345]
    MAX_RETRY = 3

    def __init__(self):
        self.processes = {}
        self.outfiles = {}
        self.cuteids = {}
        self.start_times = {}
        self.retry_counts = {}
        self.job_specs = {}
    
    def _cleanup_proc(self, pk, timeout=0):
        self._kill(pk, timeout=timeout)
        self.processes[pk].communicate()
        self.outfiles[pk].close()
        for d in (self.processes, self.outfiles, self.cuteids, self.start_times,
                  self.retry_counts, self.job_specs):
            del d[pk]
    
    def _check_retcode(self, proc, timeout):
        try: 
            retcode = proc.wait(timeout=timeout)
        except TimeoutExpired:
            retcode = None
        return retcode

    def _check_retcodes(self):
        start = time.time()
        pk_retcodes = []
        for pk, proc in self.processes.items():
            elapsed = time.time() - start
            timeout = max(0, self.CHECK_PERIOD - elapsed)
            retcode = self._check_retcode(proc, timeout)
            pk_retcodes.append((pk, retcode))
        return pk_retcodes

    def _log_error_tail(self, pk, retcode):
        fname = self.outfiles[pk].name
        if os.path.exists(fname):
            tail = get_tail(self.outfiles[pk].name)
        else:
            tail = ''
        logmsg = self.log_prefix(pk) + f'nonzero return {retcode}:\n {tail}'
        logger.error(logmsg)
        return tail
           
    def _can_retry(self, pk, retcode):
        if retcode in self.RETRY_CODES:
            elapsed = time.time() - self.start_times[pk]
            retry_count = self.retry_counts[pk]
            if elapsed < self.RETRY_WINDOW and retry_count <= self.MAX_RETRY:
                logmsg = self.log_prefix(pk) 
                logmsg += (f'can retry task (err occured after {elapsed:.2f} sec; '
                          f'attempt {self.retry_counts[pk]}/{self.MAX_RETRY})')
                logger.error(logmsg)
                return True
        return False

    def _kill(self, pk, timeout=0):
        p = self.processes[pk]
        if p.poll() is None:
            p.terminate()
            logger.debug(f"rank {RANK} sent TERM to {self.cuteids[pk]}...waiting on shutdown")
            try: p.wait(timeout=timeout)
            except TimeoutExpired: p.kill()

    def _launch_proc(self, pk):
        job_spec = self.job_specs[pk]
        workdir = job_spec['workdir']
        name = job_spec['name']
        cmd = job_spec['cmd']
        envs = job_spec['envs']
        envscript = job_spec['envscript']
        
        if envscript:
            args = ' '.join(['source', envscript, '&&', cmd])
            shell = True
        else:
            args = shlex.split(cmd)
            shell = False

        if self.gpus_per_node > 0:
            idx = list(self.job_specs.keys()).index(pk) 
            gpu_device = idx % self.gpus_per_node
            envs['CUDA_DEVICE_ORDER'] = "PCI_BUS_ID"
            envs['CUDA_VISIBLE_DEVICES'] = str(gpu_device)

        out_name = f'{name}.out'
        logger.info(f"{self.log_prefix(pk)} Popen (shell={shell}):\n{args}")

        if not os.path.exists(workdir): os.makedirs(workdir)
        outfile = open(os.path.join(workdir, out_name), 'wb')
        self.outfiles[pk] = outfile
        try:
            proc = Popen(args, stdout=outfile, stderr=STDOUT,
                         cwd=workdir, env=envs, shell=shell,)
        except Exception as e:
            proc = FailedToStartProcess()
            logger.error(self.log_prefix(pk) + f"Popen error:\n{str(e)}\n")
            sleeptime = 0.5 + 3.5*random.random()
            time.sleep(sleeptime)
        self.processes[pk] = proc
    
    def _handle_error(self, pk, retcode):
        tail = self._log_error_tail(pk, retcode)

        if not self._can_retry(pk, retcode):
            self._cleanup_proc(pk)
            return (retcode, tail)
        else:
            self.outfiles[pk].close()
            self.start_times[pk] = time.time()
            self.retry_counts[pk] += 1
            self._launch_proc(pk)
            return 'running'

    def log_prefix(self, pk=None):
        prefix = f'rank {RANK} '
        if pk: prefix += f'{self.cuteids[pk]} '
        return prefix

    def write_message(self, job_statuses):
        msg = {'ask' : [], 'done' : [], 'error': []}
        num_jobs = len(job_statuses)
        num_errors = len([
            s for s in job_statuses.values()
            if s not in ["running","done"]
        ])
        max_tail = (MSG_BUFSIZE - 110*num_jobs) // max(1,num_errors)
        for pk, status in job_statuses.items():
            if status == 'running':
                msg['ask'].append(pk)
            elif status == 'done':
                msg['done'].append(pk)
            else:
                retcode, tail = status
                tail = tail[-max_tail:]
                msg['error'].append((pk, retcode, tail))
        return msg

    def update_processes(self):
        statuses = {}
        for pk, retcode in self._check_retcodes():
            if retcode is None:
                statuses[pk] = 'running'
            elif retcode == 0:
                statuses[pk] = 'done'
                self._cleanup_proc(pk)
            else:
                statuses[pk] = self._handle_error(pk, retcode)
        return statuses
    
    def exit(self):
        all_pks = list(self.processes.keys())
        for pk in all_pks:
            self._cleanup_proc(pk, timeout=self.CHECK_PERIOD)
        MPI.Finalize()
        sys.exit(0)

    def start_jobs(self, msg):
        assert msg['tag'] == 'NEW'
        for pk in msg:
            if pk == 'tag': continue
            job_spec = msg[pk]
            self.job_specs[pk] = job_spec
            self.cuteids[pk] = job_spec['cuteid']
            self.start_times[pk] = time.time()
            self.retry_counts[pk] = 1
            self._launch_proc(pk)

    def kill_jobs(self, kill_pks):
        for pk in kill_pks: self._cleanup_proc(pk, timeout=0)

    def main(self):
        tag = None
        gpus_per_node = None
        self.gpus_per_node = comm.bcast(gpus_per_node, root=0)
        while tag != 'EXIT':
            logger.debug(f"rank {RANK} waiting on recv from master...")
            msg = comm.recv(source=0)
            tag = msg['tag']
            logger.debug(f"rank {RANK} recv done: got msg tag {tag}")

            if tag == 'NEW':
                self.start_jobs(msg)
            elif tag == 'KILL':
                self.kill_jobs(msg['kill_pks'])
            elif tag == 'EXIT':
                logger.debug(f"rank {RANK} received EXIT")
                break

            statuses = self.update_processes()
            cuteids = ' '.join(self.cuteids.values())
            logger.debug(f"rank {RANK} jobs: {cuteids}")
            if len(statuses) > 0:
                msg = self.write_message(statuses)
                logger.debug(f"rank {RANK} sending request to master...")
                comm.send(msg, dest=0)
                logger.debug(f"rank {RANK} send done")
        self.exit()

if __name__ == "__main__":
    if RANK == 0:
        master = Master()
        def handle_term(signum, stack): master.EXIT_FLAG = True
        signal.signal(signal.SIGINT, handle_term)
        signal.signal(signal.SIGTERM, handle_term)
        master.main()
    else:
        worker = Worker()
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        worker.main()



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import balsam.launcher.dag as dag
import glob
import sys
import os
current_job = dag.current_job

print("Hello from make_sides_post")

if dag.ERROR or dag.TIMEOUT:
    if dag.ERROR: 
        print("make_sides_post recognized error flag")
    else: 
        print("make_sides_post recognized timeout flag")

    num_sides = int(os.environ['BALSAM_FT_NUM_SIDES'])
    num_files = len(glob.glob("side*.dat"))

    if num_files == num_sides:
        print("it's okay, the job was actually done")
        current_job.update_state("JOB_FINISHED", "handled error; it was okay")
        exit(0)
    elif num_files == 0:
        print("Creating rescue job")
        children = current_job.get_children()
        rescue = dag.spawn_child(clone=True, application_args="--sleep 0 --retcode 0")
        rescue.set_parents([])
        current_job.update_state("JOB_FINISHED", f"spawned rescue job {rescue.cute_id}")
        for child in children:
            child.set_parents([rescue])
        exit(0)

if '--dynamic-spawn' not in sys.argv:
    sys.exit(0)

reduce_job = current_job.get_child_by_name('sum_squares')

for i, sidefile in enumerate(glob.glob("side*.dat")):
    square_job = dag.spawn_child(name=f"square{i}", application="square",
                    application_args=sidefile, input_files=sidefile)
    dag.add_dependency(parent=square_job, child=reduce_job)
    print(f"spawned square{i} job")



'''
The Balsam Launcher is a pilot application that runs either directly on compute
nodes or very close to them (perhaps on a service node with ``mpirun`` access). 
The Launcher is responsible for:

    * **pulling jobs** from the Balsam database

    * **stage-in and stage-out** transitions that include working
      directory creation, remote and local file transfer, and the enabling the
      requested flow of files from parent to child jobs transparently

    * running custom **pre- and post-processing** scripts for each job

    * invoking **job execution** on the appropriate resources

    * **monitoring** job execution and providing resilient mechanisms to **handle
      expected and unexpected** runtime errors or job timeouts

In normal Balsam usage, the Launcher is not invoked directly. Instead, multiple
Launcher instances with specific workloads are automatically scheduled for execution by the
Metascheduler.  However, the Launcher can also simply be invoked from the
command line.  For example, to consume all jobs from the database, use:

>>> $ balsam launcher --consume-all 
'''

import logging
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)



''' Scheduler Exceptions '''

class SchedulerException(Exception): pass
class SubmitNonZeroReturnCode(SchedulerException): pass
class SubmitSubprocessFailed(SchedulerException): pass
class JobSubmitFailed(SchedulerException): pass
class JobSubmissionDisabled(SchedulerException): pass
class JobStatusFailed(SchedulerException): pass
class StatusNonZeroReturnCode(SchedulerException): pass
class NoQStatInformation(SchedulerException): pass



from datetime import datetime
import logging
import logging.handlers
import os
import sys

try:
    import django
    from django.db import OperationalError
    from django.conf import settings
except ImportError:
    print("Warning: Django is not installed")

from balsam.__version__ import __version__

def setup():
    if not settings.configured:
        os.environ['DJANGO_SETTINGS_MODULE'] = 'balsam.django_config.settings'
        django.setup()

_logger = logging.getLogger()
_logger.setLevel(logging.DEBUG)
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)

def config_logging(basename):
    timestamp = datetime.now().strftime('%Y-%m-%d_%H%M%S')
    fname = f'{basename}_{timestamp}.log'
    HANDLER_FILE = os.path.join(settings.LOGGING_DIRECTORY, fname)
        
    formatter=logging.Formatter(
            '%(asctime)s|%(process)d|%(levelname)8s|%(name)s:%(lineno)s] %(message)s', 
            datefmt="%d-%b-%Y %H:%M:%S"
            )
    handler = logging.FileHandler(filename=HANDLER_FILE)
    level = getattr(logging, settings.LOG_HANDLER_LEVEL)
    handler.setLevel(level)
    handler.setFormatter(formatter)
    _logger.addHandler(handler)

def log_uncaught_exceptions(exctype, value, tb):
    if isinstance(value, OperationalError) and not settings.DATABASES['default']['PORT']:
        _logger.error("Balsam OperationalError: No DB is currently active")
        _logger.error("Please use `source balsamactivate` to activate a Balsam DB")
        _logger.error("Use `balsam which --list` for a listing of known DB names")
    else:
        _logger.error(f"Uncaught Exception {exctype}: {value}",exc_info=(exctype,value,tb))
        for handler in _logger.handlers: handler.flush()

sys.excepthook = log_uncaught_exceptions
__all__ = ['config_logging', 'settings', 'setup']



import logging
import os
import sys
import signal
import stat
import time
from socket import gethostname

from django.db.models import Count
from balsam import config_logging, settings, setup
from balsam.core import models
from balsam.service import queues, jobpacker
from balsam.service.schedulers import script_template, scheduler
from balsam.scripts.cli import service_subparser
from balsam.core import transitions

logger = logging.getLogger('balsam.service.service')
QueuedLaunch = models.QueuedLaunch
BalsamJob = models.BalsamJob
source = BalsamJob.source
EXIT_FLAG = False

def submit_qlaunch(qlaunch, verbose=False):
    top = settings.SERVICE_PATH
    pk = qlaunch.pk
    script_path = os.path.join(top, f'qlaunch{pk}.sh')
    if os.path.exists(script_path):
        raise ValueError("Job script already rendered for {qlaunch}")
    script = script_template.render(qlaunch)

    with open(script_path, 'w') as fp:
        fp.write(script)
    st = os.stat(script_path)
    os.chmod(script_path, st.st_mode | stat.S_IEXEC)
    try:
        sched_id = scheduler.submit(script_path)
    except Exception as e:
        logger.error(f'Failed to submit job for {qlaunch}:\n{e}')
        qlaunch.delete()
        raise
    else:
        qlaunch.scheduler_id = sched_id
        qlaunch.state = "submitted"
        qlaunch.command = script_path
        qlaunch.save(update_fields=['scheduler_id','state','command'])
        msg = f'Submit OK: {qlaunch}'
        logger.info(msg)
        if verbose: print(msg)

def sig_handler(signum, stack):
    global EXIT_FLAG
    EXIT_FLAG = True

def get_open_queues():
    query = QueuedLaunch.objects.values('queue').annotate(num_queued=Count('queue'))
    num_queued = {d['queue'] : d['num_queued'] for d in query}
    open_queues = {qname:queue for qname,queue in queues.queues.items()
                   if num_queued.get(qname,0) < queue['max_queued']}
    return open_queues

def main(args):
    signal.signal(signal.SIGINT,  sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGHUP, signal.SIG_IGN)

    if not QueuedLaunch.acquire_advisory():
        print("Another service is currently running on this Balsam DB")
        sys.exit(1)

    while not EXIT_FLAG:
        QueuedLaunch.refresh_from_scheduler()
        open_queues = get_open_queues()
        if open_queues:
            logger.info(f"Open queues: {list(open_queues.keys())}")
            qlaunch = jobpacker.create_qlaunch(open_queues)
            if qlaunch: submit_qlaunch(qlaunch)
        if not QueuedLaunch.acquire_advisory():
            logger.error('Failed to refresh advisory lock; aborting')
            break
        elif not EXIT_FLAG:
            source.clear_stale_locks()
            time.sleep(10)

if __name__ == "__main__":
    setup()
    config_logging('service')
    logger.info(f"Balsam Service starting on {gethostname()}")
    parser = service_subparser()
    transition_pool = transitions.TransitionProcessPool(1, '')
    source.start_tick()
    try:
        main(parser.parse_args())
    except: 
        raise
    finally: 
        transition_pool.terminate()
        source.release_all_owned()
        logger.info(f"Balsam Service shutdown: released all locks OK")



import shlex
import subprocess
from balsam.service.schedulers.exceptions import *
import logging
logger = logging.getLogger(__name__)

class Scheduler:
    SCHEDULER_VARIABLES = {}
    JOBSTATUS_VARIABLES = {}

    def __init__(self):
        logger.debug(f"Using scheduler class {self.__class__}")

    def submit(self, script_path):
        submit_cmd = self._make_submit_cmd(script_path)
        p = subprocess.run(submit_cmd, stdout=subprocess.PIPE,shell=True,
                             stderr=subprocess.STDOUT, encoding='utf-8')
        if p.returncode != 0: 
            raise SubmitNonZeroReturnCode(p.stdout)
        scheduler_id = self._parse_submit_output(p.stdout)
        return scheduler_id

    def _status(self):
        stat_cmd = self._make_status_cmd()
        p = subprocess.run(stat_cmd, stdout=subprocess.PIPE, shell=True,
                           stderr=subprocess.STDOUT, encoding='utf-8')
        if p.returncode != 0: 
            raise StatusNonZeroReturnCode(p.stdout)
        statinfo = self._parse_status_output(p.stdout)
        return statinfo

    def get_status(self, scheduler_id):
        scheduler_id = int(scheduler_id)
        try:
            statuses = self._status()
        except StatusNonZeroReturnCode as e:
            raise NoQStatInformation("QStat failed: {e}")
        try:
            stat = statuses[scheduler_id]
        except KeyError:
            raise NoQStatInformation(f"No status for {scheduler_id}: this might signal the job is over")
        else:
            return stat

    def status_dict(self):
        return self._status()



from django.conf import settings
import configparser
import json
import os
from collections import namedtuple
import logging
import pprint
logger = logging.getLogger('balsam.service')

Queue = namedtuple('Queue', ['name', 'min_time', 'max_time'])

class _QueuePolicy:
    def __init__(self):
        self.queues = {}
        self.max_nodes = 0
        self.min_nodes = 0
        
        top = settings.BALSAM_HOME
        policy_fname = settings.QUEUE_POLICY
        policypath = os.path.join(top, policy_fname)
        config = configparser.ConfigParser()
        config.read(policypath)
        for queue_name in config.sections(): 
            qconf = config[queue_name]
            self.add_from_config(queue_name, qconf)
        msg = pprint.pformat(self.queues, indent=4)
        logger.info('Loaded Queue Policy\n'+msg)

    def add_from_config(self, queue_name, qconf):
        global_max_nodes = self.max_nodes
        global_min_nodes = self.min_nodes
        try:
            submit_jobs = qconf.getboolean('submit-jobs')
            max_queued = qconf.getint('max-queued')
            qpolicy = json.loads(qconf['policy'])
        except Exception as e:
            logger.error(f'Failed to parse {queue_name}: {e}')
            return
        if not submit_jobs or max_queued==0: return
        try:
            q = {'max_queued' : max_queued}
            for rule in qpolicy:
                min_nodes = int(rule['min-nodes'])
                max_nodes = int(rule['max-nodes'])
                assert min_nodes <= max_nodes
                assert (min_nodes, max_nodes) not in q
                global_max_nodes = max(global_max_nodes, max_nodes)
                global_min_nodes = min(global_min_nodes, min_nodes)
                min_time = float(rule['min-time'])
                max_time = float(rule['max-time'])
                assert min_time <= max_time
                q[(min_nodes, max_nodes)] = (min_time, max_time)
        except Exception as e:
            logger.error(f'Failed to parse policy for {queue_name}: {e}')
        else:
            self.queues[queue_name] = q
            self.max_nodes = global_max_nodes
            self.min_nodes = global_min_nodes

    def find_queue(self, open_queues, num_nodes):
        for qname in open_queues:
            queue = self.queues[qname]
            for node_range, time_range in queue.items():
                if node_range == 'max_queued': continue
                low, high = node_range
                if low <= num_nodes <= high:
                    min_time, max_time = time_range
                    return Queue(name, min_time, max_time)

_queue_policy = _QueuePolicy()
max_nodes = _queue_policy.max_nodes
min_nodes = _queue_policy.min_nodes
queues = _queue_policy.queues
find_queue = _queue_policy.find_queue



import sys,os,subprocess,shlex,logging,time,random
from django.conf import settings
logger = logging.getLogger(__name__)

# job has completed, M,M,M,copy files from gridftp server
#logger.debug('setup X509 certificates for gridftp calls')
#os.environ['X509_USER_CERT'] = '/users/hpcusers/balsam/gridsecurity/jchilders/xrootdsrv-cert.pem'
#os.environ['X509_USER_KEY']  = '/users/hpcusers/balsam/gridsecurity/jchilders/xrootdsrv-key.pem'

RETRY_ATTEMPTS=10

def globus_url_copy(from_path,to_path):

   p = subprocess.Popen([settings.GRIDFTP_PROXY_INFO,'-exists'])
   p.wait()
   if p.returncode is not 0: # valid proxy does not exist so create one
      p = subprocess.Popen([settings.GRIDFTP_PROXY_INIT,'-verify','-debug','-bits','2048','-valid','96:00'],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
      stdout = ''
      for line in p.stdout:
         logger.debug(line[0:-1])
         stdout += line
      p.wait()
      if p.returncode is not 0:
         raise Exception('grid-proxy-init failed: stdout = \n' + stdout + '\n')
      else:
         logger.debug(' grid-proxy-init stdout: \n' + stdout + '\n')


   from_path = str(from_path)
   to_path = str(to_path)
   cmd = settings.GRIDFTP_GLOBUS_URL_COPY + ' -cd  -r -nodcau ' + from_path + ' ' + to_path
   #logger.debug(' GridFtp, from_path = ' + from_path )
   #logger.debug(' GridFtp, to_path   = ' + to_path )
   
   # added random sleep to keep too many grid ftp requests from hammering the server at the same time
   time.sleep(random.randint(1,100))

   logger.debug(' transferring files: ' + cmd )
   for i in range(RETRY_ATTEMPTS):
      try:
         proc = subprocess.Popen(shlex.split(cmd),stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
      except:
         logger.error(' Error while performing GridFTP transfer, retrying transfer ' + str(i+1) + ' of ' + str(RETRY_ATTEMPTS) )
         if i+1 >= RETRY_ATTEMPTS: raise
         else:
            logger.error(' waiting 10 minutes before trying transfer again ')
            time.sleep(600)
            continue
      stdout = ''
      for line in proc.stdout:
         logger.debug(line[0:-1])
         stdout += line
      p.wait()
      if p.returncode is not 0:
         logger.error(' globus-url-copy exited with non-zero return code: stdout = \n' + stdout + '\n')
         if i+1 == RETRY_ATTEMPTS:
            raise Exception('globus-url-copy: exited with non-zero return code: returncode = ' + str(p.returncode) + '\n  stdout = \n' + stdout + '\n')
         else:
            logger.error(' waiting 10 minutes before trying transfer again ')
            time.sleep(600)
            continue
      elif 'error' in stdout.lower():
         logger.error('globus-url-copy: error found in output: stdout = \n' + stdout + '\n')
         if i+1 == RETRY_ATTEMPTS:
            raise Exception('globus-url-copy: error found in output: stdout = \n' + stdout + '\n')
         else:
            logger.error(' waiting 10 minutes before trying transfer again ')
            time.sleep(600)
            continue
      
   return stdout





'''BalsamJob Transitions

The user selects ``NUM_TRANSITION_THREADS`` processes to run
alongside the main Launcher process.  These are created with ``multiprocessing.Process``
and communicate with the Launcher through a ``multiprocessing.Queue`` (or a
PriorityQueue via a ``multiprocessing.managers.SyncManager``)

The transition processes pull jobs from the queue, execute the necessary
transition function, and signal completion by putting a status message back on a
status queue. These transition processes explicitly ignore SIGINT and SIGTERM
interrupts, so that they can finish the current transition and exit gracefully,
under control of the main Launcher process.
'''
from collections import defaultdict
import glob
import multiprocessing
import os
from io import StringIO
from traceback import print_exc
import random
import signal
import shutil
import subprocess
import time
import tempfile

from django import db
from django.db.models.functions import Cast, Substr
from django.db.models import CharField

from balsam.core import transfer 
from balsam.launcher.exceptions import *
try:
    from balsam.core.models import BalsamJob
except:
    from balsam.launcher import dag
    BalsamJob = dag.BalsamJob

from balsam.core.models import PROCESSABLE_STATES, END_STATES
from balsam.launcher.util import get_tail

import logging
logger = logging.getLogger(__name__)

JOBCACHE_LIMIT = 1000
PREPROCESS_TIMEOUT_SECONDS = 300
POSTPROCESS_TIMEOUT_SECONDS = 300
EXIT_FLAG = False

def handler(signum, stack):
    global EXIT_FLAG
    EXIT_FLAG = True

class TransitionProcessPool:
    '''Launch and terminate the transition processes'''
    def __init__(self, num_threads, wf_name):

        self.procs = [
            multiprocessing.Process(
                target=main,
                args=(i, num_threads, wf_name)
            )
            for i in range(num_threads)
        ]
        logger.info(f"Starting {len(self.procs)} transition processes")
        db.connections.close_all()
        for proc in self.procs:
            proc.daemon = True
            proc.start()

    def terminate(self):
        '''Terminate workers via signal and process join'''
        logger.debug("Sending sigterm and waiting on transition processes")
        for proc in self.procs:
            proc.terminate()
        for proc in self.procs: 
            proc.join()
        logger.info("All Transition processes joined: done.")

@db.transaction.atomic
def fail_update(failed_jobs):
    for job in failed_jobs:
        try: failmsg = job.__fail_msg
        except AttributeError: failmsg = ''
        job.refresh_from_db()
        job.update_state('FAILED', failmsg)

def update_states_from_cache(job_cache):
    # Update states of fast-forwarded jobs
    update_jobs = defaultdict(list)
    failed_jobs = []
    for job in job_cache:
        if job.state != job.__old_state:
            job.__old_state = job.state
            if job.state != 'FAILED': update_jobs[job.state].append(job.pk)
            else: failed_jobs.append(job)

    if failed_jobs: fail_update(failed_jobs)
    for newstate, joblist in update_jobs.items():
        BalsamJob.batch_update_state(joblist, newstate)

def select_range(num_threads, thread_idx):
    HEX_DIGITS = '0123456789abcdef'
    chunk, rem = divmod(len(HEX_DIGITS), num_threads)
    start, end = thread_idx*chunk, (thread_idx+1)*chunk
    my_digits = HEX_DIGITS[start:end]
    if thread_idx < rem:
        my_digits += HEX_DIGITS[thread_idx-rem]

    manager = BalsamJob.source
    processable = manager.by_states(PROCESSABLE_STATES).filter(lock='')
    processable = processable.order_by('-state') # put AWAITING_PARENTS last to avoid starvation
    logger.debug(f"There are {processable.count()} processable jobs")
    if num_threads == 1:
        qs = processable
    else:
        qs = processable.annotate(first_pk_char=Substr(Cast('pk', CharField(max_length=36)) , 1, 1))
        qs = qs.filter(first_pk_char__in=my_digits)
    qs = qs.values_list('pk', flat=True)[:JOBCACHE_LIMIT]
    logger.debug(f"TransitionThread{thread_idx} select:\n{qs.query}")
    return list(qs)

def refresh_cache(job_cache, num_threads, thread_idx):
    manager = BalsamJob.source
    to_acquire = select_range(num_threads, thread_idx)
    logger.debug(f"TransitionThread{thread_idx} will try to acquire: {[str(id)[:8] for id in to_acquire]}")
    acquired = manager.acquire(to_acquire)

    if len(acquired) <  len(to_acquire):
        st = random.random()
        time.sleep(st)
        logger.debug(f"failed to acquire {len(to_acquire)}; only got {len(acquired)}")
    if acquired:
        logger.debug(f'Acquired {len(acquired)} new jobs')
        acquired = BalsamJob.objects.filter(pk__in=acquired)
        job_cache.extend(acquired)
    for job in job_cache: job.__old_state = job.state

def release_jobs(job_cache):
    manager = BalsamJob.source
    release_jobs = [
        j.pk for j in job_cache if
        (j.state not in PROCESSABLE_STATES)
        or (j.state == 'AWAITING_PARENTS')
    ]
    if release_jobs: manager.release(release_jobs)
    return [j for j in job_cache if j.pk not in release_jobs]

def main(thread_idx, num_threads, wf_name):
    global EXIT_FLAG
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)
    
    manager = BalsamJob.source
    manager.workflow = wf_name
    random.seed(multiprocessing.current_process().pid)
    time.sleep(random.random())
    manager.start_tick()

    try:
        _main(thread_idx, num_threads)
    except:
        buf = StringIO()
        print_exc(file=buf)
        logger.critical(f"Uncaught exception:\n%s", buf.getvalue())
    finally:
        manager.release_all_owned()
        logger.debug('Transition process finished: released all jobs')

def _main(thread_idx, num_threads):
    global EXIT_FLAG
    manager = BalsamJob.source
    job_cache = []
    last_refresh = 0
    refresh_period = 5

    while not EXIT_FLAG:
        # Update in-memory cache of locked BalsamJobs
        elapsed = time.time() - last_refresh
        if elapsed > refresh_period:
            if len(job_cache) < JOBCACHE_LIMIT:
                refresh_cache(job_cache, num_threads, thread_idx)
            last_refresh = time.time()
        else:
            time.sleep(1)

        # Fast-forward transitions & release locks
        fast_forward(job_cache)
        job_cache = release_jobs(job_cache)

        # Run transitions (one pass over all jobs)
        for job in job_cache:
            transition_function = TRANSITIONS[job.state]
            try:
                transition_function(job)
            except BalsamTransitionError as e:
                job.state = 'FAILED'
                buf = StringIO()
                print_exc(file=buf)
                job.__fail_msg = buf.getvalue()
                logger.exception(f"Marking {job.cute_id} as FAILED")
            if EXIT_FLAG:
                break
        # Update states in bulk
        update_states_from_cache(job_cache)
        job_cache = release_jobs(job_cache)
    logger.info('EXIT_FLAG: exiting main loop')


def check_parents(job):
    '''Check job's dependencies, update to READY if satisfied'''
    num_parents = len(job.get_parents_by_id())

    if num_parents == 0 or not job.wait_for_parents:
        ready = True
    else:
        parents = job.get_parents()
        ready = num_parents == parents.filter(state='JOB_FINISHED').count()

    if ready:
        job.state = 'READY'
        logger.debug(f'{job.cute_id} ready')
    elif job.state != 'AWAITING_PARENTS':
        job.state = 'AWAITING_PARENTS'
        logger.info(f'{job.cute_id} waiting for {num_parents} parents')
    elif parents.filter(state__in=['FAILED', 'USER_KILLED']).exists():
        job.state = 'FAILED'
        job.__fail_msg = 'One or more parent jobs failed'

def fast_forward(job_cache):
    '''Make several passes over the job list; advancing states in order'''
    # Check parents
    check_jobs = (j for j in job_cache if j.state in 'CREATED AWAITING_PARENTS'.split())
    for job in check_jobs: check_parents(job)

    # Skip stage-in
    stagein_jobs = (j for j in job_cache if j.state == 'READY')
    for job in stagein_jobs:
        workdir = job.working_directory
        if not os.path.exists(workdir):
            os.makedirs(workdir)
            logger.info(f"{job.cute_id} created working directory {workdir}")
        hasParents = bool(job.get_parents_by_id())
        hasInput = bool(job.input_files)
        hasRemote = bool(job.stage_in_url)
        if not hasRemote and not (hasParents and hasInput): job.state = 'STAGED_IN'

    # Skip preprocess
    preprocess_jobs = (j for j in job_cache if j.state == 'STAGED_IN')
    for job in preprocess_jobs:
        if not job.preprocess: job.state = 'PREPROCESSED'

    # RUN_DONE: skip postprocess
    done_jobs = (j for j in job_cache if j.state=='RUN_DONE' and not j.postprocess)
    for job in done_jobs: job.state = 'POSTPROCESSED'

    # Timeout: retry
    retry_jobs = (j for j in job_cache if j.state=='RUN_TIMEOUT' and j.auto_timeout_retry and not j.post_timeout_handler)
    for job in retry_jobs: job.state = 'RESTART_READY'
    
    # Timeout: fail
    timefail_jobs = (j for j in job_cache if j.state=='RUN_TIMEOUT'
                     and not j.auto_timeout_retry
                     and not (j.postprocess and j.post_timeout_handler)
                    )
    for job in timefail_jobs: job.state = 'FAILED'

    # Error: fail 
    errfail_jobs = (j for j in job_cache if j.state=='RUN_ERROR'
                    and not (j.post_error_handler and j.postprocess)
                   )
    for job in errfail_jobs: job.state = 'FAILED'

    # skip stageout (finished)
    stageout_jobs = (j for j in job_cache if j.state=='POSTPROCESSED'
                     and not (j.stage_out_url and j.stage_out_files)
                    )
    for job in stageout_jobs: job.state = 'JOB_FINISHED'
    update_states_from_cache(job_cache)


def stage_in(job):
    logger.debug(f'{job.cute_id} in stage_in')

    work_dir = job.working_directory
    if not os.path.exists(work_dir):
        os.makedirs(workdir)
        logger.debug(f"{job.cute_id} working directory {work_dir}")

    # stage in all remote urls
    # TODO: stage_in remote transfer should allow a list of files and folders,
    # rather than copying just one entire folder
    url_in = job.stage_in_url
    if url_in:
        logger.info(f"{job.cute_id} transfer in from {url_in}")
        try:
            transfer.stage_in(f"{url_in}",  f"{work_dir}")
        except Exception as e:
            message = 'Exception received during stage_in: ' + str(e)
            raise BalsamTransitionError(message) from e

    # create unique symlinks to "input_files" patterns from parents
    # TODO: handle data flow from remote sites transparently
    matches = []
    parents = job.get_parents()
    input_patterns = job.input_files.split()
    logger.debug(f"{job.cute_id} searching parent workdirs for {input_patterns}")
    for parent in parents:
        parent_dir = parent.working_directory
        for pattern in input_patterns:
            path = os.path.join(parent_dir, pattern)
            matches.extend((parent.pk,match) 
                           for match in glob.glob(path))

    for parent_pk, inp_file in matches:
        basename = os.path.basename(inp_file)
        new_path = os.path.join(work_dir, basename)
        
        if os.path.exists(new_path): new_path += f"_{str(parent_pk)[:8]}"
        # pointing to src, named dst
        logger.info(f"{job.cute_id}   {new_path}  -->  {inp_file}")
        try:
            os.symlink(src=inp_file, dst=new_path)
        except FileExistsError:
            logger.warning(f"Symlink at {new_path} already exists; skipping creation")
        except Exception as e:
            raise BalsamTransitionError(
                f"Exception received during symlink: {e}") from e

    job.state = 'STAGED_IN'
    logger.debug(f"{job.cute_id} stage_in done")


def stage_out(job):
    '''copy from the local working_directory to the output_url '''
    logger.debug(f'{job.cute_id} in stage_out')

    url_out = job.stage_out_url
    if not url_out:
        job.state = 'JOB_FINISHED'
        logger.debug(f'{job.cute_id} no stage_out_url: done')
        return

    stage_out_patterns = job.stage_out_files.split()
    logger.debug(f"{job.cute_id} stage out files match: {stage_out_patterns}")
    work_dir = job.working_directory
    matches = []
    for pattern in stage_out_patterns:
        path = os.path.join(work_dir, pattern)
        matches.extend(glob.glob(path))

    if matches:
        logger.info(f"{job.cute_id} stage out files: {matches}")
        with tempfile.TemporaryDirectory() as stagingdir:
            try:
                for f in matches: 
                    base = os.path.basename(f)
                    dst = os.path.join(stagingdir, base)
                    shutil.copyfile(src=f, dst=dst)
                    logger.info(f"staging {f} out for transfer")
                logger.info(f"transferring to {url_out}")
                transfer.stage_out(f"{stagingdir}/*", f"{url_out}/")
            except Exception as e:
                message = f'Exception received during stage_out: {e}'
                raise BalsamTransitionError(message) from e
    job.state = 'JOB_FINISHED'
    logger.debug(f'{job.cute_id} stage_out done')


def preprocess(job):
    logger.debug(f'{job.cute_id} in preprocess')

    # Get preprocesser exe
    preproc_app = job.preprocess
    if not preproc_app:
        job.state = 'PREPROCESSED'
        return

    if not os.path.exists(preproc_app.split()[0]):
        #TODO: look for preproc in the EXE directories
        message = f"Preprocessor {preproc_app} does not exist on filesystem"
        raise BalsamTransitionError(message)

    # Create preprocess-specific environment
    envs = job.get_envs()

    # Run preprocesser with special environment in job working directory
    out = os.path.join(job.working_directory, f"preprocess.log")
    with open(out, 'w') as fp:
        fp.write(f"# Balsam Preprocessor: {preproc_app}")
        fp.flush()
        try:
            args = preproc_app.split()
            logger.info(f"{job.cute_id} preprocess Popen {args}")
            proc = subprocess.Popen(args, stdout=fp,
                                    stderr=subprocess.STDOUT, env=envs,
                                    cwd=job.working_directory,
                                    )
            retcode = proc.wait(timeout=PREPROCESS_TIMEOUT_SECONDS)
            proc.communicate()
        except Exception as e:
            message = f"Preprocess failed: {e}"
            try: proc.kill()
            except: pass
            raise BalsamTransitionError(message) from e

    if retcode != 0:
        tail = get_tail(out)
        message = f"{job.cute_id} preprocess returned {retcode}:\n{tail}"
        raise BalsamTransitionError(message)

    job.state = 'PREPROCESSED'
    logger.debug(f"{job.cute_id} preprocess done")

def postprocess(job, *, error_handling=False, timeout_handling=False):
    logger.debug(f'{job.cute_id} in postprocess')
    if error_handling and timeout_handling:
        raise ValueError("Both error-handling and timeout-handling is invalid")
    if error_handling: logger.info(f'{job.cute_id} handling RUN_ERROR')
    if timeout_handling: logger.info(f'{job.cute_id} handling RUN_TIMEOUT')

    # Get postprocesser exe
    postproc_app = job.postprocess

    # If no postprocesssor; move on (unless in error_handling mode)
    if not postproc_app:
        if error_handling:
            message = f"{job.cute_id} handle error: no postprocessor found!"
            raise BalsamTransitionError(message)
        elif timeout_handling:
            job.state = 'RESTART_READY'
            logger.warning(f'{job.cute_id} unhandled job timeout: marked RESTART_READY')
            return
        else:
            job.state = 'POSTPROCESSED',
            logger.debug(f'{job.cute_id} no postprocess: skipped')
            return

    if not os.path.exists(postproc_app.split()[0]):
        #TODO: look for postproc in the EXE directories
        message = f"Postprocessor {postproc_app} does not exist on filesystem"
        raise BalsamTransitionError(message)

    # Create postprocess-specific environment
    envs = job.get_envs(timeout=timeout_handling, error=error_handling)

    # Run postprocesser with special environment in job working directory
    out = os.path.join(job.working_directory, f"postprocess.log")
    with open(out, 'w') as fp:
        fp.write(f"# Balsam Postprocessor: {postproc_app}\n")
        if timeout_handling: fp.write("# Invoked to handle RUN_TIMEOUT\n")
        if error_handling: fp.write("# Invoked to handle RUN_ERROR\n")
        fp.flush()
        
        try:
            args = postproc_app.split()
            logger.info(f"{job.cute_id} postprocess Popen {args}")
            proc = subprocess.Popen(args, stdout=fp,
                                    stderr=subprocess.STDOUT, env=envs,
                                    cwd=job.working_directory,
                                    )
            retcode = proc.wait(timeout=POSTPROCESS_TIMEOUT_SECONDS)
            proc.communicate()
        except Exception as e:
            message = f"Postprocess failed: {e}"
            try: proc.kill()
            except: pass
            raise BalsamTransitionError(message) from e
    
    if retcode != 0:
        tail = get_tail(out, nlines=30)
        message = f"{job.cute_id} postprocess returned {retcode}:\n{tail}"
        raise BalsamTransitionError(message)

    job.refresh_from_db()
    # If postprocessor handled error or timeout, it should have changed job's
    # state. If it failed to do this, mark FAILED.  Otherwise, POSTPROCESSED.
    if error_handling and job.state == 'RUN_ERROR':
        message = f"{job.cute_id} Error handling didn't fix job state: marking FAILED"
        raise BalsamTransitionError(message)

    if timeout_handling and job.state == 'RUN_TIMEOUT':
        message = f"{job.cute_id} Timeout handling didn't change job state: marking FAILED"
        raise BalsamTransitionError(message)

    # Only move the state along to POSTPROCESSED if the job is still in RUN_DONE
    # and the post.py returned normally.  Otherwise, post.py might mark a job
    # FAILED, and you override it with POSTPROCESSED, breaking the workflow.
    if job.state == 'RUN_DONE':
        job.state = 'POSTPROCESSED'
    logger.debug(f"{job.cute_id} postprocess done")


def handle_timeout(job):
    if job.post_timeout_handler:
        logger.debug(f'{job.cute_id} invoking postprocess with timeout_handling flag')
        postprocess(job, timeout_handling=True)
    else:
        raise BalsamTransitionError(f"{job.cute_id} no timeout handling: marking FAILED")


def handle_run_error(job):
    if job.post_error_handler:
        logger.debug(f'{job.cute_id} invoking postprocess with error_handling flag')
        postprocess(job, error_handling=True)
    else:
        raise BalsamTransitionError("No error handler: run failed")

TRANSITIONS = {
    'AWAITING_PARENTS': check_parents,
    'READY':            stage_in,
    'STAGED_IN':        preprocess,
    'RUN_DONE':         postprocess,
    'RUN_TIMEOUT':      handle_timeout,
    'RUN_ERROR':        handle_run_error,
    'POSTPROCESSED':    stage_out,
}



from datetime import datetime
from importlib.util import find_spec

from balsam.service.models import BalsamJob
from tests.BalsamTestCase import BalsamTestCase
from tests.BalsamTestCase import create_job, create_app

from tests import util
    
class TestNoOp(BalsamTestCase):

    def setUp(self):
        from itertools import product

        self.launcherInfo = util.launcher_info()
        max_workers = self.launcherInfo.num_workers

        num_nodes = [2**n for n in range(0,13) if 2**n <= max_workers]
        if num_nodes[-1] != max_workers:
            num_nodes.append(max_workers)

        rpn = [64]
        jpn = [64, 512]
        #jpn = [16]
        self.experiments = product(num_nodes, rpn, jpn)

    def create_serial_expt(self, num_nodes, rpn, jpn):
        '''Populate DB with set number of dummy serial jobs; no deps'''
        BalsamJob.objects.all().delete()
        num_jobs = num_nodes * jpn
        
        jobs = [create_job(name=f'task{i}', direct_command=f'echo Hello',
                         args=str(i), workflow='bench-no-op', save=False)
                         for i in range(num_jobs)]
        BalsamJob.objects.bulk_create(jobs)
        self.assertEqual(BalsamJob.objects.count(), num_jobs)

    def test_serial(self):
        '''Populate DB, run launcher, get timing data from job histories
        Serial: all jobs pack into MPIEnsembles and can run concurrently'''
        done_query = BalsamJob.objects.filter(state='JOB_FINISHED')

        for (num_nodes, rpn, jpn) in self.experiments:
            title = f'{num_nodes}nodes_{rpn}rpn_{jpn}jpn'
            self.create_serial_expt(num_nodes, rpn, jpn)
            num_jobs = num_nodes * jpn

            launcher_start_time = datetime.now()
            success = util.run_launcher_until(lambda: done_query.count() == num_jobs, 
                                              timeout=1000, maxrpn=rpn)
            self.assertEqual(done_query.count(), num_jobs)

            time_data = util.process_job_times(time0=launcher_start_time)
            self.assertEqual(len(time_data['PREPROCESSED']), num_jobs)
            self.assertEqual(len(time_data['JOB_FINISHED']), num_jobs)

            cdf_table = util.print_jobtimes_cdf(time_data)
            resultpath = util.benchmark_outfile_path('serial_no_op.dat')

            with open(resultpath, 'w') as fp:
                title = f'# {num_nodes} nodes, {rpn} rpn, {jpn} jpn ({num_jobs} total jobs)'
                comment = 'All jobs pack into MPIEnsembles and can run concurrently'
                fp.write(util.FormatTable.create_header(title, comment))
                fp.write(cdf_table)



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import argparse
import time
import sys

print("Hello from square")
parser = argparse.ArgumentParser()
parser.add_argument('infile')
parser.add_argument('--sleep', type=float, default=0)
parser.add_argument('--retcode', type=int, default=0)
args = parser.parse_args()

side_length = float(open(args.infile).read())
with open('square.dat', 'w') as fp:
    square = side_length**2
    fp.write(str(square) + "\n")

time.sleep(args.sleep)
sys.exit(args.retcode)



from django.conf import settings
from importlib import import_module
from balsam.service.schedulers import JobEnvironment, JobTemplate

_schedClass = settings.SCHEDULER_CLASS.strip()
_temp = import_module('balsam.service.schedulers.'+_schedClass)
scheduler = _temp.new_scheduler()
JobEnv = JobEnvironment.JobEnvironment(scheduler)

template_path = settings.BALSAM_HOME
template_name = settings.JOB_TEMPLATE
script_template = JobTemplate.ScriptTemplate(template_path, template_name)

__all__ = ['scheduler', 'JobEnv', 'script_template']



from django.shortcuts import render, redirect
from balsam.core.models import BalsamJob
from django.http import JsonResponse, HttpResponseBadRequest

def list_tasks(request):
    if not request.is_ajax():
        return HttpResponseBadRequest()
    
    #print("Parsed GET dict:")
    #print('\n'.join(f"{p} == {request.GET[p]}" for p in request.GET))

    _jobs = BalsamJob.objects.values_list('job_id', 'name', 'workflow',
                                          'application', 'state',
                                          'queued_launch__scheduler_id',
                                          'num_nodes', 'ranks_per_node')
    return JsonResponse({"data": list(_jobs)})

def task_detail(request, id):
    task = BalsamJob.objects.get(pk=id)
    text = repr(task)
    result = '<table cellpadding="5" cellspacing="0" border="0" style="padding-left:50px;">\n'
    lines = text.split('\n')[2:-1]
    for line in lines:
        try: field, val = line.split(':', 2)
        except: continue
        else: result += f'<tr> <td> {field} </td> <td> {val} </td> </tr>\n'
    result += '</table>'
    return JsonResponse({"html": result})



from collections import defaultdict, Counter
from itertools import combinations
import os
import json
import logging
import re
import numpy as np
import socket
import sys
import threading
from datetime import datetime, timedelta
from django.utils import timezone
import uuid
from getpass import getuser
from pprint import pformat

from balsam import setup
setup()

from django.core.exceptions import ValidationError,ObjectDoesNotExist
from django.db.utils import OperationalError
from django.conf import settings
from django.db import models, transaction
from django.db.models import Value as V
from django.db.models import Q
from django.db import connection
from django.db.models.functions import Concat
from django.contrib.postgres.fields import JSONField

logger = logging.getLogger(__name__)

class InvalidStateError(ValidationError): pass
class InvalidParentsError(ValidationError): pass
class NoApplication(Exception): pass

TIME_FMT = '%m-%d-%Y %H:%M:%S.%f'

STATES = '''
CREATED
AWAITING_PARENTS
READY

STAGED_IN
PREPROCESSED

RUNNING
RUN_DONE

POSTPROCESSED
JOB_FINISHED

RUN_TIMEOUT
RUN_ERROR
RESTART_READY

FAILED
USER_KILLED
'''.split()

ACTIVE_STATES = '''
RUNNING
'''.split()

PROCESSABLE_STATES = '''
CREATED
AWAITING_PARENTS
READY
STAGED_IN
RUN_DONE
POSTPROCESSED
RUN_TIMEOUT
RUN_ERROR
'''.split()

RUNNABLE_STATES = '''
PREPROCESSED
RESTART_READY
'''.split()

END_STATES = '''
JOB_FINISHED
FAILED
USER_KILLED
'''.split()
        
STATE_TIME_PATTERN = re.compile(r'''
^                  # start of line
\[                 # opening square bracket
(\d+-\d+-\d\d\d\d  # date MM-DD-YYYY
\s+                # one or more space
\d+:\d+:\d+\.\d+)  # time HH:MM:SS.MICROSEC
\s+                # one or more space
(\w+)              # state
\s*                # 0 or more space
\]                 # closing square bracket
''', re.VERBOSE | re.MULTILINE)

_app_cache = {}

def process_job_times(qs=None):
    '''Returns {state : [elapsed_seconds_for_each_job_to_reach_state]}
    Useful for tracking job performance/throughput'''

    if qs is None: qs = BalsamJob.objects
    data = qs.values_list('state_history', flat=True)
    data = '\n'.join(data)
    matches = STATE_TIME_PATTERN.finditer(data)
    result = ( m.groups() for m in matches )
    result = ( (state, datetime.strptime(time_str, TIME_FMT))
              for (time_str, state) in result )
    time_data = defaultdict(list)
    for state, time in result:
        time_data[state].append(time)
    return time_data

def utilization_report(time_data=None):
    if time_data is None:
        qs = BalsamJob.objects
        time_data = process_job_times(qs=qs)
    start_times = time_data.get('RUNNING', [])
    end_times = []
    for state in ['RUN_DONE', 'RUN_ERROR', 'RUN_TIMEOUT']:
        end_times.extend(time_data.get(state, []))

    startCounts = Counter(start_times)
    endCounts = Counter(end_times)
    for t in endCounts: endCounts[t] *= -1
    merged = sorted(list(startCounts.items()) + list(endCounts.items()),
                    key = lambda x: x[0])
    counts = np.fromiter((x[1] for x in merged), dtype=np.int)

    times = [x[0] for x in merged]
    running = np.cumsum(counts)
    return (times, running)

def throughput_report(time_data=None):
    if time_data is None:
        qs = BalsamJob.objects
        time_data = process_job_times(qs=qs)
    done_times = time_data.get('RUN_DONE', [])
    doneCounts = sorted(list(Counter(done_times).items()),key=lambda x:x[0])
    times = [x[0] for x in doneCounts]
    counts = np.cumsum(np.fromiter((x[1] for x in doneCounts), dtype=np.int))
    return (times, counts)

def error_report(time_data=None):
    if time_data is None:
        qs = BalsamJob.objects
        time_data = process_job_times(qs=qs)
    err_times = time_data.get('RUN_ERROR', [])
    if not err_times: return
    time0 = min(err_times)
    err_seconds = np.array([(t-time0).total_seconds() for t in err_times])
    hmin, hmax = 0, max(err_seconds)
    bins = np.arange(hmin, hmax+60, 60)
    times = [time0 + timedelta(seconds=s) for s in bins]
    hist, _ = np.histogram(err_seconds, bins=bins, density=False)
    assert len(times) == len(hist) + 1
    return times, hist

def assert_disjoint():
    groups = [ACTIVE_STATES, PROCESSABLE_STATES, RUNNABLE_STATES, END_STATES]
    joined = [state for g in groups for state in g]
    assert len(joined) == len(set(joined)) == len(STATES)
    assert set(joined) == set(STATES) 
    for g1,g2 in combinations(groups, 2):
        s1,s2 = set(g1), set(g2)
        assert s1.intersection(s2) == set()
assert_disjoint()

def validate_state(value):
    if value not in STATES:
        raise InvalidStateError(f"{value} is not a valid state in balsam.models")

def get_time_string():
    return timezone.now().strftime(TIME_FMT)

def from_time_string(s):
    return datetime.strptime(s, TIME_FMT)

def history_line(state='CREATED', message=''):
    return f"\n[{get_time_string()} {state}] ".rjust(46) + message

def safe_select(queryset):
    qs = queryset.order_by('job_id').select_for_update()
    pks = None
    while pks is None:
        try:
            pks = list(qs.values_list('job_id', flat=True))
        except OperationalError as e:
            logger.error(f'select for update error: {e}\nRetrying...')
            time.sleep(0.2)
    return BalsamJob.objects.filter(pk__in=pks)


class QueuedLaunch(models.Model):

    ADVISORY_LOCK_ID = hash(getuser())
    scheduler_id = models.IntegerField(default=0, db_index=True)
    project = models.TextField(default=settings.DEFAULT_PROJECT)
    queue = models.TextField(default='')
    nodes = models.IntegerField(default=0)
    wall_minutes = models.IntegerField(default=0)
    job_mode = models.TextField(default='')
    wf_filter = models.TextField(default='')
    command = models.TextField(default='')
    state = models.TextField(default='pending-submission')
    prescheduled_only = models.BooleanField(default=True) # if disabled, all BalsamJobs eligible to run
    from_balsam = models.BooleanField(default=True) # if disabled, all BalsamJobs eligible to run

    @classmethod
    def acquire_advisory(self):
        with connection.cursor() as cursor:
            command = f"SELECT pg_try_advisory_lock({self.ADVISORY_LOCK_ID})"
            cursor.execute(command)
            row = cursor.fetchone()
        row = ' '.join(map(str, row)).strip().lower()
        if 'true' in row:
            return True
        else:
            return False

    def __repr__(self):
        dat = {k:v for k,v in self.__dict__.items() if k not in ['_state']}
        return f'''Qlaunch {pformat(dat, indent=4)}'''

    def __str__(self):
        return repr(self)

    @classmethod
    @transaction.atomic
    def refresh_from_scheduler(cls):
        from balsam.service.schedulers import scheduler
        saved_jobs = list(cls.objects.all())
        saved_job_ids = [j.scheduler_id for j in saved_jobs]
        stats = scheduler.status_dict()
        for job_id, job in stats.items():
            if job_id not in saved_job_ids:
                j = cls(scheduler_id=job_id,
                    project=job['project'],
                    queue=job['queue'],
                    nodes=job['nodes'],
                    wall_minutes=job['wall_time_min'],
                    state=job['state'],
                    command=job['command'],
                    from_balsam=False)
                j.save()
                logger.info(f'Detected new job: {j}')
            else:
                saved_job = saved_jobs[saved_job_ids.index(job_id)]
                if job['state'] != saved_job.state: 
                    logger.info(f'Updating batch job {job_id}: state {job["state"]}')
                saved_job.state = job['state']
                saved_job.project = job['project']
                saved_job.queue = job['queue']
                saved_job.nodes = job['nodes']
                saved_job.wall_minutes = job['wall_time_min']
                saved_job.command = job['command']
                saved_job.save()
        delete_ids = [id for id in saved_job_ids if id not in stats.keys()]
        cls.objects.filter(scheduler_id__in=delete_ids).delete()
        if delete_ids:
            logger.info(f'Deleting Jobs {delete_ids} no longer in scheduler')

class JobSource(models.Manager):

    TICK_PERIOD = timedelta(minutes=1)
    EXPIRATION_PERIOD = timedelta(minutes=3)

    def __init__(self, workflow=None):
        super().__init__()
        self.workflow = workflow
        self._lock_str = None
        self._pid = None
        self.qLaunch = None
        self._checked_qLaunch = False

    def check_qLaunch(self):
        from balsam.service.schedulers import JobEnv
        sched_id = JobEnv.current_scheduler_id
        if sched_id is not None:
            try:
                self.qLaunch = QueuedLaunch.objects.get(scheduler_id=sched_id)
            except ObjectDoesNotExist:
                self.qLaunch = None
        if self.qLaunch is not None:
            if not (self.qLaunch.prescheduled_only and self.qLaunch.from_balsam):
                self.qLaunch = None
        if self.qLaunch is not None:
            logger.info(f'Filtering BalsamJobs scheduled for QueuedLaunch {self.qLaunch.pk} (Batch Job: {self.qLaunch.scheduler_id})')
        else:
            logger.info(f'Job source filtering for un-scheduled BalsamJobs')
        self._checked_qLaunch = True

    @property
    def lock_str(self):
        pid = os.getpid()
        if pid != self._pid:
            self._lock_str = f"{socket.gethostname()}:{pid}"
            self._pid = pid
        return self._lock_str

    @property
    def lockQuery(self):
        return Q(lock='') | Q(lock=self.lock_str)

    def get_queryset(self):
        if not self._checked_qLaunch: self.check_qLaunch()

        queryset = super().get_queryset()
        queryset = queryset.filter(self.lockQuery)
        if self.qLaunch is not None:
            queryset = queryset.filter(queued_launch_id=self.qLaunch.pk)
        else:
            queryset = queryset.filter(queued_launch__isnull=True)
        if self.workflow:
            queryset = queryset.filter(workflow__contains=self.workflow)
        return queryset

    def by_states(self, states):
        if isinstance(states, str):
            states = [states]
        elif isinstance(states, dict):
            states = states.keys()
        return self.get_queryset().filter(state__in=states)

    def get_runnable(self, *, max_nodes, remaining_minutes=None, mpi_only=False,
                     serial_only=False, order_by=None):
        if mpi_only and serial_only:
            raise ValueError("arguments mpi_only and serial_only are mutually exclusive")

        if max_nodes < 1:
            raise ValueError("Must be positive number of nodes")

        if serial_only:
            assert max_nodes == 1

        runnable = self.by_states(RUNNABLE_STATES)
        runnable = runnable.filter(num_nodes__lte=max_nodes)

        if remaining_minutes is not None:
            try: remaining_minutes = int(remaining_minutes)
            except: remaining_minutes = None
            else: runnable = runnable.filter(wall_time_minutes__lte=remaining_minutes)

        if serial_only:
            runnable = runnable.filter(num_nodes=1, ranks_per_node=1)
        elif mpi_only:
            mpiquery = Q(num_nodes__gt=1) | Q(ranks_per_node__gt=1)
            runnable = runnable.filter(mpiquery)
        if order_by is not None:
            if isinstance(order_by, str):
                order_by = (order_by,)
            runnable = runnable.order_by(*order_by)
        return runnable

    @transaction.atomic
    def acquire(self, pk_list):
        '''input can be actual list of PKs or a queryset'''
        new_lock = self.lock_str
        to_lock = BalsamJob.objects.filter(pk__in=pk_list)
        to_lock = to_lock.select_for_update(skip_locked=True).filter(lock='')
        acquired_pks = list(to_lock.values_list('job_id', flat=True))
        BalsamJob.objects.filter(pk__in=acquired_pks).update(lock=new_lock,
                tick=timezone.now())
        return acquired_pks

    def start_tick(self):
        t = threading.Timer(self.TICK_PERIOD.total_seconds(), self.start_tick)
        t.daemon = True
        t.start()
        self._tick()

    def _tick(self):
        now = timezone.now()
        with transaction.atomic():
            my_locked = safe_select(BalsamJob.objects.filter(lock=self.lock_str))
            num_updated = my_locked.update(tick=now)
        logger.debug(f'Ticked lock on {num_updated} of my BalsamJobs')
        connection.close()

    def release(self, pk_list):
        with transaction.atomic():
            to_unlock = safe_select(BalsamJob.objects.filter(pk__in=pk_list))
            num_unlocked = to_unlock.update(lock='')
        logger.debug(f'Released lock on {num_unlocked} of my BalsamJobs')

    @transaction.atomic
    def release_all_owned(self):
        alljobs = safe_select(BalsamJob.objects.filter(lock=self.lock_str))
        alljobs.update(lock='')
    
    def clear_stale_locks(self):
        objects = self.model.objects
        total_count = objects.count()
        locked_count = objects.exclude(lock='').count()
        logger.debug(f'{locked_count} out of {total_count} jobs are locked')

        all_jobs = objects.all()
        expired_time = timezone.now() - self.EXPIRATION_PERIOD
        with transaction.atomic():
            expired_jobs = safe_select(all_jobs.exclude(lock='').filter(tick__lte=expired_time))
            revert_count = expired_jobs.filter(state='RUNNING').update(state='RESTART_READY')
            expired_count = expired_jobs.update(lock='')
        if expired_count:
            logger.info(f'Cleared stale lock on {expired_count} jobs')
            if revert_count: logger.info(f'Reverted {revert_count} RUNNING jobs to RESTART_READY')
        elif locked_count:
            logger.debug(f'No stale locks (older than {self.EXPIRATION_PERIOD.total_seconds()} seconds)')

class BalsamJob(models.Model):
    ''' A DB representation of a Balsam Job '''

    objects = models.Manager()
    source = JobSource()

    job_id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False)

    workflow = models.TextField(
        'Workflow Name',
        help_text='Name of the workflow to which this job belongs',
        default='')
    name = models.TextField(
        'Job Name',
        help_text='A name for the job given by the user.',
        default='')
    description = models.TextField(
        'Job Description',
        help_text='A description of the job.',
        default='')
    lock = models.TextField(
        'Process Lock',
        help_text='{hostname}:{PID} set by process that currently owns the job',
        default='',
        db_index=True
    )
    tick = models.DateTimeField(auto_now_add=True)

    parents = models.TextField(
        'IDs of the parent jobs which must complete prior to the start of this job.',
        default='[]')

    input_files = models.TextField(
        'Input File Patterns',
        help_text="Space-delimited filename patterns that will be searched in the parents'"\
        "working directories. Every matching file will be made available in this"\
        "job's working directory (symlinks for local Balsam jobs, file transfer for"\
        "remote Balsam jobs). Default: all files from parent jobs are made available.",
        default='*')
    stage_in_url = models.TextField(
        'External stage in files or folders', help_text="A list of URLs for external data to be staged in prior to job processing. Job dataflow from parents to children is NOT handled here; see `input_files` field instead.",
        default='')
    stage_out_files = models.TextField(
        'External stage out files or folders',
        help_text="A string of filename patterns. Matches will be transferred to the stage_out_url. Default: no files are staged out",
        default='')
    stage_out_url = models.TextField(
        'Stage Out URL',
        help_text='The URLs to which designated stage out files are sent.',
        default='')

    wall_time_minutes = models.IntegerField(
        'Job Wall Time in Minutes',
        help_text='The number of minutes the job is expected to take',
        default=1)
    num_nodes = models.IntegerField(
        'Number of Compute Nodes',
        help_text='The number of compute nodes requested for this job.',
        default=1,
        db_index=True)
    coschedule_num_nodes = models.IntegerField(
        'Number of additional compute nodes to reserve alongside this job',
        help_text='''Used by Balsam service only.  If a pilot job runs on one or a
        few nodes, but requires additional worker nodes alongside it,
        use this field to specify the number of additional nodes that will be
        reserved by the service for this job.''',
        default=0)
    ranks_per_node = models.IntegerField(
        'Number of ranks per node',
        help_text='The number of MPI ranks per node to schedule for this job.',
        default=1)
    cpu_affinity = models.TextField(
        'Cray CPU Affinity ("depth" or "none")',
        default="none")
    threads_per_rank = models.IntegerField(
        'Number of threads per MPI rank',
        help_text='The number of OpenMP threads per MPI rank (if applicable)',
        default=1)
    threads_per_core = models.IntegerField(
        'Number of hyperthreads per physical core (if applicable)',
        help_text='Number of hyperthreads per physical core.',
        default=1)
    node_packing_count = models.IntegerField(
        'For serial (non-MPI) jobs only. How many to run concurrently on a node.',
        help_text='Setting this field at 2 means two serial jobs will run at a '
        'time on a node. This field is ignored for MPI jobs.',
        default=1)
    environ_vars = models.TextField(
        'Environment variables specific to this job',
        help_text="Colon-separated list of envs like VAR1=value1:VAR2=value2",
        default='')
    
    application = models.TextField(
        'Application to Run',
        help_text='The application to run; located in Applications database',
        default='')
    args = models.TextField(
        'Command-line args to the application exe',
        help_text='Command line arguments used by the Balsam job runner',
        default='')
    user_workdir = models.TextField(
        'Override the Balsam-generated workdir, point to existing location',
        default=''
    )


    wait_for_parents = models.BooleanField(
            'If True, do not process this job until parents are FINISHED',
            default=True)
    post_error_handler = models.BooleanField(
        'Let postprocesser try to handle RUN_ERROR',
        help_text='If true, the postprocessor will be invoked for RUN_ERROR jobs'
        ' and it is up to the script to handle error and update job state.',
        default=False)
    post_timeout_handler = models.BooleanField(
        'Let postprocesser try to handle RUN_TIMEOUT',
        help_text='If true, the postprocessor will be invoked for RUN_TIMEOUT jobs'
        ' and it is up to the script to handle timeout and update job state.',
        default=False)
    auto_timeout_retry = models.BooleanField(
        'Automatically restart jobs that have timed out',
        help_text="If True and post_timeout_handler is False, then jobs will "
        "simply be marked RESTART_READY upon timing out.",
        default=True)

    state = models.TextField(
        'Job State',
        help_text='The current state of the job.',
        default='CREATED',
        validators=[validate_state],
        db_index=True)
    state_history = models.TextField(
        'Job State History',
        help_text="Chronological record of the job's states",
        default=history_line)

    queued_launch = models.ForeignKey(
        'QueuedLaunch',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    data = JSONField('User Data', help_text="JSON encoded data store for user-defined data", default=dict)

    @staticmethod
    def from_dict(d):
        job = BalsamJob()
        SERIAL_FIELDS = [f for f in job.__dict__ if f not in
                '_state force_insert force_update using update_fields'.split()
                ]

        if type(d['job_id']) is str:
            d['job_id'] = uuid.UUID(d['job_id'])
        else:
            assert d['job_id'] is None
            d['job_id'] = job.job_id

        for field in SERIAL_FIELDS:
            job.__dict__[field] = d[field]

        assert type(job.job_id) == uuid.UUID
        return job


    def __repr__(self):
        result = f'BalsamJob {self.pk}\n'
        result += '----------------------------------------------\n'
        result += '\n'.join( (k+':').ljust(32) + str(v) 
                for k,v in self.__dict__.items() 
                if k not in ['state_history', 'job_id', '_state', 'tick'])

        try: result += '\n' + '  *** Executed command:'.ljust(32) + self.app_cmd
        except NoApplication: result += '\n' + '  *** Executed command:'.ljust(32) + f'NO APPLICATION MATCHING {self.application}'
        except ApplicationDefinition.DoesNotExist: result += '\n' + '  *** Executed command:'.ljust(32) + f'NO APPLICATION MATCHING {self.application}'

        result += '\n' + '  *** Working directory:'.ljust(32) + self.working_directory +'\n'
        return result

    def __str__(self):
        return self.__repr__()

    def get_parents_by_id(self):
        return json.loads(self.parents)

    def get_parents(self):
        parent_ids = self.get_parents_by_id()
        return BalsamJob.objects.filter(job_id__in=parent_ids)

    @property
    def num_ranks(self):
        return self.num_nodes * self.ranks_per_node

    @property
    def cute_id(self):
        if self.name:
            return f"[{self.name} | { str(self.pk)[:8] }]"
        else:
            return f"[{ str(self.pk)[:8] }]"
    
    @property
    def app_cmd(self):
        app = self.get_application()
        line = f"{app.executable} {self.args}"
        return ' '.join(os.path.expanduser(w) for w in line.split())

    @property
    def envscript(self):
        app = self.get_application()
        _envscript = app.envscript
        if _envscript and os.path.isfile(_envscript):
            return _envscript
        else:
            return None


    def get_children(self):
        return BalsamJob.objects.filter(parents__icontains=str(self.pk))

    def get_children_by_id(self):
        children = self.get_children()
        return [c.pk for c in children]

    def get_child_by_name(self, name):
        children = self.get_children().filter(name=name)
        if children.count() == 0:
            raise ValueError(f"No child named {name}")
        elif children.count() > 1:
            raise ValueError(f"More than one child named {name}")
        else:
            return children.first()

    def set_parents(self, parents):
        try:
            parents_list = list(parents)
        except:
            raise InvalidParentsError("Cannot convert input to list")
        for i, parent in enumerate(parents_list):
            pk = parent.pk if isinstance(parent,BalsamJob) else parent
            if not BalsamJob.objects.filter(pk=pk).exists():
                raise InvalidParentsError(f"Job PK {pk} is not in the BalsamJob DB")
            parents_list[i] = str(pk)
        self.parents = json.dumps(parents_list)
        self.save(update_fields=['parents'])

    def get_application(self):
        if not self.application: 
            raise NoApplication
        elif self.application in _app_cache:
            return _app_cache[self.application]
        else:
            app = ApplicationDefinition.objects.get(name=self.application)
            _app_cache[self.application] = app
            return app

    @property
    def preprocess(self):
        try:
            app = self.get_application()
            return app.preprocess
        except NoApplication:
            return ''
    
    @property
    def postprocess(self):
        try:
            app = self.get_application()
            return app.postprocess
        except NoApplication:
            return ''

    @staticmethod
    def parse_envstring(s):
        result = {}
        entries = s.split(':')
        entries = [e.split('=') for e in entries]
        return {variable:'='.join(values) for (variable,*values) in entries}

    def get_envs(self, *, timeout=False, error=False):
        envs = os.environ.copy()
        
        if self.environ_vars:
            job_vars = self.parse_envstring(self.environ_vars)
            envs.update(job_vars)
    
        balsam_envs = dict(
            BALSAM_JOB_ID=str(self.pk),
            BALSAM_PARENT_IDS=str(self.parents),
        )

        if self.threads_per_rank > 1:
            balsam_envs['OMP_NUM_THREADS'] = str(self.threads_per_rank)

        if timeout: balsam_envs['BALSAM_JOB_TIMEOUT']="TRUE"
        if error: balsam_envs['BALSAM_JOB_ERROR']="TRUE"
        envs.update(balsam_envs)
        return envs

    @classmethod
    def batch_update_state(cls, pk_list, new_state, message=''):
        try:
            exists = pk_list.exists()
        except AttributeError:
            exists = bool(pk_list)
        if not exists: return

        if new_state not in STATES:
            raise InvalidStateError(f"{new_state} is not a job state in balsam.models")
        
        msg = history_line(new_state, message)

        with transaction.atomic():
            update_jobs = cls.objects.filter(job_id__in=pk_list).exclude(state='USER_KILLED')
            update_jobs = safe_select(update_jobs)
            update_jobs.update(state=new_state,
                               state_history=Concat('state_history', V(msg))
                              )

    def update_state(self, new_state, message=''):
        if new_state not in STATES:
            raise InvalidStateError(f"{new_state} is not a job state in balsam.models")
        msg = history_line(new_state, message)
        self.state = new_state
        self.state_history += msg
        self.save(update_fields=['state', 'state_history'])

    def get_recent_state_str(self):
        return self.state_history.split("\n")[-1].strip()

    def read_file_in_workdir(self, fname):
        work_dir = self.working_directory
        path = os.path.join(work_dir, fname)
        if not os.path.exists(path):
            raise ValueError(f"{fname} not found in working directory of {self.cute_id}")
        else:
            return open(path).read()

    def get_state_times(self):
        matches = STATE_TIME_PATTERN.findall(self.state_history)
        return {state: datetime.strptime(timestr, TIME_FMT)
                for timestr, state in matches
               }

    @property
    def runtime_seconds(self):
        times = self.get_state_times()
        t0 = times.get('RUNNING', None) 
        t1 = times.get('RUN_DONE', None) 
        if t0 and t1:
            return (t1-t0).total_seconds()
        else:
            return None

    @property
    def working_directory(self):
        if self.user_workdir and os.path.isdir(self.user_workdir):
            return self.user_workdir
        top = settings.BALSAM_WORK_DIRECTORY
        if self.workflow:
            top = os.path.join(top, self.workflow)
        name = self.name.strip().replace(' ', '_')
        name += '_' + str(self.pk)[:8]
        path = os.path.join(top, name)
        return path

    def to_dict(self):
        SERIAL_FIELDS = [f for f in self.__dict__ if f not in ['_state']]
        d = {field : self.__dict__[field] for field in SERIAL_FIELDS}
        return d

    def serialize(self, **kwargs):
        d = self.to_dict()
        d.update(kwargs)
        if type(self.job_id) == uuid.UUID:
            d['job_id'] = str(self.job_id)
        else:
            assert self.job_id == d['job_id'] == None

        serial_data = json.dumps(d)
        return serial_data

    @classmethod
    def deserialize(cls, serial_data):
        if type(serial_data) is bytes:
            serial_data = serial_data.decode('utf-8')
        if type(serial_data) is str:
            serial_data = json.loads(serial_data)
        job = BalsamJob.from_dict(serial_data)
        return job

class ApplicationDefinition(models.Model):
    ''' application definition, each DB entry is a task that can be run
        on the local resource. '''
    name = models.TextField(
        'Application Name',
        help_text='The name of an application that can be run locally.',
        default='')
    description = models.TextField(
        'Application Description',
        help_text='A description of the application.',
        default='')
    executable = models.TextField(
        'Executable',
        help_text='The executable path to run this application on the local system.',
        default='')
    preprocess = models.TextField(
        'Preprocessing Script',
        help_text='A script that is run in a job working directory prior to submitting the job to the queue.',
        default='')
    envscript = models.TextField(
        'Environment Setup Script',
        help_text='A script that is sourced immediately prior to the job launch command.',
        default='')
    postprocess = models.TextField(
        'Postprocessing Script',
        help_text='A script that is run in a job working directory after the job has completed.',
        default='')

    def __repr__(self):
        result = f'Application {self.pk}:\n'
        result += '-----------------------\n'
        result += '\n'.join( (k+':').ljust(32) + str(v) 
                for k,v in self.__dict__.items() 
                if k not in ['_state', 'id'])
        return result
    
    def __str__(self):
        return self.__repr__()

    @property
    def cute_id(self):
        return f"[{self.name} | { str(self.pk)[:8] }]"




import getpass
import os
from importlib.util import find_spec
import subprocess
import signal
import sys

import django

def ls_procs(keywords):
    if type(keywords) == str: keywords = [keywords]

    username = getpass.getuser()
    
    searchcmd = 'ps aux | grep '
    searchcmd += ' | grep '.join(f'"{k}"' for k in keywords) 
    grep = subprocess.Popen(searchcmd, shell=True, stdout=subprocess.PIPE)
    stdout,stderr = grep.communicate()
    stdout = stdout.decode('utf-8')

    processes = [line for line in stdout.split('\n') if 'python' in line and line.split()[0]==username]
    return processes

def cmd_confirmation(message=''):
    confirm = ''
    while not confirm.lower() in ['y', 'n']:
        try:
            confirm = input(f"{message} [y/n]: ")
        except: pass
    return confirm.lower() == 'y'

def newapp(args):
    from balsam import setup
    setup()
    from balsam.core.models import ApplicationDefinition as AppDef

    def py_app_path(path):
        if not path: 
            return ''
        args = path.split()
        app = args[0]
        if app.endswith('.py'):
            exe = sys.executable
            script_path = os.path.abspath(app)
            args = ' '.join(args[1:])
            path = ' '.join((exe, script_path, args))
        return path

    if AppDef.objects.filter(name=args.name).exists():
        raise RuntimeError(f"An application named {args.name} already exists")
    
    app = AppDef()
    app.name = args.name
    app.description = ' '.join(args.description) if args.description else ''
    app.executable = py_app_path(args.executable)
    app.preprocess = py_app_path(args.preprocess)
    app.postprocess = py_app_path(args.postprocess)
    app.save()
    print(app)
    print("Added app to database")


def newjob(args):
    from balsam import setup
    setup()
    from balsam.core import models
    Job = models.BalsamJob
    AppDef = models.ApplicationDefinition

    if not AppDef.objects.filter(name=args.application).exists():
        raise RuntimeError(f"App {args.application} not registered in local DB")

    job = Job()
    job.name = args.name
    job.description = ' '.join(args.description)
    job.workflow = args.workflow

    job.wall_time_minutes = args.wall_time_minutes
    job.num_nodes = args.num_nodes
    job.coschedule_num_nodes = args.coschedule_num_nodes
    job.node_packing_count = args.node_packing_count
    job.ranks_per_node = args.ranks_per_node
    job.threads_per_rank = args.threads_per_rank
    job.threads_per_core = args.threads_per_core

    job.application = args.application
    job.args = ' '.join(args.args)
    job.post_error_handler = args.post_handle_error
    job.post_timeout_handler = args.post_handle_timeout
    job.auto_timeout_retry = not args.disable_auto_timeout_retry
    job.input_files = ' '.join(args.input_files)

    job.stage_in_url = args.url_in
    job.stage_out_url = args.url_out
    job.stage_out_files = ' '.join(args.stage_out_files)
    job.environ_vars = ":".join(args.env)

    print(job)
    if not args.yes:
        if not cmd_confirmation('Confirm adding job to DB'):
            print("Add job aborted")
            return
    job.save()
    return job
    print("Added job to database")


def match_uniq_job(s):
    from balsam import setup
    setup()
    from balsam.core import models
    Job = models.BalsamJob

    job = Job.objects.filter(job_id__icontains=s)
    count = job.count()
    if count > 1:
        raise ValueError(f"More than one ID matched {s}")
    elif count == 1:
        return job.first()
    else:
        raise ValueError(f"No job in local DB matched {s}")

def newdep(args):
    from balsam import setup
    setup()
    from balsam.core import models
    from balsam.launcher import dag
    Job = models.BalsamJob
    AppDef = models.ApplicationDefinition

    parent = match_uniq_job(args.parent)
    child = match_uniq_job(args.child)
    dag.add_dependency(parent, child)
    print(f"Created link {parent.cute_id} --> {child.cute_id}")

def ls(args):
    from balsam import setup
    setup()
    from balsam.core import models
    from balsam.launcher import dag
    import balsam.scripts.ls_commands as lscmd
    Job = models.BalsamJob
    AppDef = models.ApplicationDefinition

    objects = args.objects
    name = args.name
    history = args.history
    verbose = args.verbose
    state = args.state
    id = args.id
    tree = args.tree
    wf = args.wf
    by_states = args.by_states

    try:
        if objects.startswith('job'):
            lscmd.ls_jobs(name, history, id, verbose, tree, wf, state, by_states)
        elif objects.startswith('app'):
            lscmd.ls_apps(name, id, verbose)
        elif objects.startswith('work') or objects.startswith('wf'):
            lscmd.ls_wf(name, verbose, tree, wf)
        elif objects.startswith('queues'):
            lscmd.ls_queues(verbose)
    except (KeyboardInterrupt,BrokenPipeError):
        pass

def modify(args):
    from balsam import setup
    setup()
    from balsam.core import models
    Job = models.BalsamJob
    AppDef = models.ApplicationDefinition

    if args.type == 'jobs': cls = Job
    elif args.type == 'apps': cls = AppDef

    item = cls.objects.filter(pk__contains=args.id)
    if item.count() == 0:
        raise RuntimeError(f"no matching {args.type}")
    elif item.count() > 1:
        raise RuntimeError(f"more than one matching {args.type}")
    item = item.first()

    target_type = type(getattr(item, args.attr))
    new_value = target_type(args.value)
    if args.attr == 'state':
        if item.state == 'USER_KILLED':
            print("Cannot mutate state of a killed job")
            return
        item.update_state(new_value, 'User mutated state from command line')
    else:
        setattr(item, args.attr, new_value)
        item.save()
    print(f'{args.type[:-1]} {args.attr} changed to:  {new_value}')


def rm(args):
    from balsam import setup
    setup()
    from balsam.core import models
    from balsam.launcher import dag
    Job = models.BalsamJob
    AppDef = models.ApplicationDefinition

    objects_name = args.objects
    name = args.name
    objid = args.id
    deleteall = args.all
    force = args.force

    # Are we removing jobs or apps?
    if objects_name.startswith('job'): cls = Job
    elif objects_name.startswith('app'): cls = AppDef
    objects = cls.objects

    # Filter: all objects, by name-match (multiple), or by ID (unique)?
    if deleteall:
        deletion_objs = objects.all()
        message = f"ALL {objects_name}"
    elif name: 
        deletion_objs = objects.filter(name__icontains=name)
        message = f"{len(deletion_objs)} {objects_name} matching name {name}"
        if not deletion_objs.exists(): 
            print("No {objects_name} matching query")
            return
    elif objid: 
        deletion_objs = objects.filter(pk__icontains=objid)
        if deletion_objs.count() > 1:
            raise RuntimeError(f"Multiple {objects_name} match ID")
        elif deletion_objs.count() == 0:
            raise RuntimeError(f"No {objects_name} match ID")
        else:
            message = f"{objects_name[:-1]} with ID matching {objid}"
    
    # User confirmation
    if not force:
        if not cmd_confirmation(f"PERMANENTLY remove {message}?"):
            print("Delete aborted")
            return

    # Actually delete things here
    deletion_objs.delete()
    print("Deleted.")

def kill(args):
    from balsam import setup
    setup()
    from balsam.core import models
    from balsam.launcher import dag
    Job = models.BalsamJob

    job_id = args.id
    
    job = Job.objects.filter(job_id__startswith=job_id)
    if job.count() > 1:
        raise RuntimeError(f"More than one job matches {job_id}")
    if job.count() == 0:
        print(f"No jobs match the given ID {job_id}")

    job = job.first()

    if cmd_confirmation(f'Really kill job {job.name} {job.cute_id} ??'):
        dag.kill(job, recursive=args.recursive)
        print("Job killed")


def mkchild(args):
    from balsam import setup
    setup()
    from balsam.launcher import dag

    if not dag.current_job:
        raise RuntimeError(f"mkchild requires that BALSAM_JOB_ID is in the environment")
    child_job = newjob(args)
    dag.add_dependency(dag.current_job, child_job)
    print(f"Created link {dag.current_job.cute_id} --> {child_job.cute_id}")

def launcher(args):
    fname = find_spec("balsam.launcher.launcher").origin
    original_args = sys.argv[2:]
    command = [sys.executable] + [fname] + original_args
    p = subprocess.Popen(command)
    print(f"Started Balsam launcher [{p.pid}]")
    p.wait()

def submitlaunch(args):
    from balsam import setup
    setup()
    from balsam.service import service
    from balsam.core import models
    from django.db import connection, transaction

    # Exclusive Lock on core_queuedlaunch
    with transaction.atomic():
        with connection.cursor() as cursor:
            cursor.execute('LOCK TABLE core_queuedlaunch IN ACCESS EXCLUSIVE MODE;')
            QueuedLaunch = models.QueuedLaunch
            qlaunch = QueuedLaunch(
                    project = args.project,
                    queue = args.queue,
                    nodes = args.nodes,
                    wall_minutes = args.time_minutes,
                    job_mode = args.job_mode,
                    wf_filter = args.wf_filter,
                    prescheduled_only=False)
            qlaunch.save()
            service.submit_qlaunch(qlaunch, verbose=True)

def service(args):
    fname = find_spec("balsam.service.service").origin
    original_args = sys.argv[2:]
    command = [sys.executable] + [fname] + original_args
    p = subprocess.Popen(command)
    print(f"Starting Balsam service [{p.pid}]")

def init(args):
    path = os.path.expanduser(args.path)
    if os.path.exists(path):
        if not os.path.isdir(path):
            print(f"{path} is not a directory")
            sys.exit(1)
    else:
        try:
            os.mkdir(path, mode=0o755)
        except:
            print(f"Failed to create directory {path}")
            sys.exit(1)
        
    fname = find_spec("balsam.scripts.init").origin
    p = subprocess.Popen(f'{sys.executable} {fname} {path}', shell=True)
    p.wait()

def which(args):
    from balsam.django_config.db_index import refresh_db_index
    from balsam.django_config.serverinfo import ServerInfo
    import pprint

    if args.list:
        pprint.pprint(refresh_db_index())
    elif args.name:
        db_list = refresh_db_index()

        # Try exact match first
        name = os.path.abspath(os.path.expanduser(args.name))
        if name in db_list:
            print(name)
            sys.exit(0)

        matches = [db for db in db_list if args.name in db]
        if len(matches) == 0:
            sys.stderr.write(f"No DB matching {args.name} is cached\n\n")
            sys.exit(1)
        elif len(matches) > 1:
            sys.stderr.write(f"DB name {args.name} is too ambiguous; multiple matches\n"
                              "Please give a longer unique substring\n\n"
                            )
            sys.exit(1)
        else:
            print(matches[0])
            sys.exit(0)
    else:
        db_path = os.environ.get('BALSAM_DB_PATH')
        if db_path:
            print("Current Balsam DB:", db_path)
            dat = ServerInfo(os.environ['BALSAM_DB_PATH']).data
            pprint.pprint(dat)
        else:
            print("BALSAM_DB_PATH is not set")
            print('Use "source balsamactivate" to activate one of these existing databases:')
            pprint.pprint(refresh_db_index())

def log(args):
    from balsam import settings, setup
    setup()
    path = os.path.join(settings.LOGGING_DIRECTORY, '*.log')
    try: subprocess.run(f"tail -f {path}", shell=True)
    except (KeyboardInterrupt,BrokenPipeError,ProcessLookupError): pass


def server(args):
    from balsam.scripts import postgres_control
    db_path = os.environ.get('BALSAM_DB_PATH', None)
    if not db_path:
        raise RuntimeError('BALSAM_DB_PATH needs to be set before server can be started\n')

    if args.connect:
        postgres_control.start_main(db_path)
    elif args.reset:
        postgres_control.reset_main(db_path)
    elif args.list_active_connections:
        postgres_control.list_connections(db_path)
    elif args.add_user:
        uname = args.add_user
        if len(uname) == 0: raise RuntimeError("Please provide user name")
        postgres_control.add_user(db_path, uname)
    elif args.drop_user:
        uname = args.drop_user
        if len(uname) == 0: raise RuntimeError("Please provide user name")
        postgres_control.drop_user(db_path, uname)
    elif args.list_users:
        postgres_control.list_users(db_path)

def make_dummies(args):
    from balsam import setup
    setup()
    from balsam.core import models
    Job = models.BalsamJob
    App = models.ApplicationDefinition
    if not App.objects.filter(name='dummy').exists():
        dummy_app = App(name="dummy", executable="echo")
        dummy_app.save()

    jobs = [Job(
                name = f'dummy{i}',
                description = 'Added by balsam make_dummies',
                node_packing_count = 64,
                workflow = 'dummy',
                application = 'dummy',
                args = 'hello'
               )
            for i in range(args.num)]
    Job.objects.bulk_create(jobs)
    print(f"Added {args.num} dummy jobs to the DB")



'''The MPICommand subclasses express a template for system-specfic MPI calls. If
the Launcher detects a specific host type, the appropriate MPICommand class is
automatically used.  Otherwise, the DEFAULTMPICommand class is assigned at
module-load time, testing for MPICH or OpenMPI'''

import subprocess
import sys
import logging
from balsam.service.schedulers import JobEnv
logger = logging.getLogger(__name__)

class MPICommand(object):
    '''Base Class for creating ``mpirun`` command lines.

    System-specific commands are generated by subclasses that specify the
    command and argument names. MPICommand instances are callable; the relevant
    parameters are passed in as arguments and an MPI command is generated from a
    template and returned'''
    def __init__(self):
        self.mpi = ''
        self.nproc = ''
        self.ppn = ''
        self.env = ''
        self.cpu_binding = None
        self.threads_per_rank = None
        self.threads_per_core = None

    def worker_str(self, workers):
        return ""

    def env_str(self, envs):
        envstrs = (f'{self.env} {var}="{val}"' for var,val in envs.items())
        return " ".join(envstrs)

    def threads(self, cpu_affinity, thread_per_rank, thread_per_core):
        return ""


    def __call__(self, workers, *, app_cmd, num_ranks, ranks_per_node,
                 envs, cpu_affinity, threads_per_rank=1,threads_per_core=1):
        '''Build the mpirun/aprun/runjob command line string'''
        workers = self.worker_str(workers)
        envs = self.env_str(envs)
        thread_str = self.threads(cpu_affinity, threads_per_rank, threads_per_core)
        result =  (f"{self.mpi} {self.nproc} {num_ranks} {self.ppn} "
                   f"{ranks_per_node} {envs} {workers} {thread_str} {app_cmd}")
        return result

class OPENMPICommand(MPICommand):
    '''Single node OpenMPI: ppn == num_ranks'''
    def __init__(self):
        self.mpi = 'mpirun'
        self.nproc = '-n'
        self.ppn = '-npernode'
        self.env = '-x'
        self.cpu_binding = None
        self.threads_per_rank = None
        self.threads_per_core = None

    def worker_str(self, workers):
        return ""

    def env_str(self, envs):
        envstrs = (f'{self.env} {var}="{val}"' for var,val in envs.items())
        return " ".join(envstrs)

    def threads(self, cpu_affinity, thread_per_rank, thread_per_core):
        return ""

    def __call__(self, workers, *, app_cmd, num_ranks, ranks_per_node, envs, cpu_affinity, threads_per_rank=1,threads_per_core=1):
        '''Build the mpirun/aprun/runjob command line string'''
        workers = self.worker_str(workers)
        envs = self.env_str(envs)
        thread_str = self.threads(cpu_affinity, threads_per_rank, threads_per_core)
        result =  (f"{self.mpi} {self.nproc} {num_ranks} {self.ppn} "
                   f"{ranks_per_node} {envs} {workers} {thread_str} {app_cmd}")
        return result


class BGQMPICommand(MPICommand):
    def __init__(self):
        self.mpi = 'runjob'
        self.nproc = '--np'
        self.ppn = '-p'
        self.env = '--envs' # VAR1=val1:VAR2=val2
        self.cpu_binding = None
        self.threads_per_rank = None
        self.threads_per_core = None
    
    def worker_str(self, workers):
        if len(workers) != 1:
            raise BalsamRunnerException("BGQ requires exactly 1 worker (sub-block)")
        worker = workers[0]
        shape, block, corner = worker.shape, worker.block, worker.corner
        return f"--shape {shape} --block {block} --corner {corner} "

class THETAMPICommand(MPICommand):
    def __init__(self):
        # 64 independent jobs, 1 per core of a KNL node: -n64 -N64 -d1 -j1
        self.mpi = 'aprun'
        self.nproc = '-n'
        self.ppn = '-N'
        self.env = '-e'
        self.cpu_binding = '-cc'
        self.threads_per_rank = '-d'
        self.threads_per_core = '-j'
    
    def threads(self, affinity, thread_per_rank, thread_per_core):
        assert affinity in 'depth none'.split()
        result = f"{self.cpu_binding} {affinity} "
        if affinity == 'depth':
            assert thread_per_rank >= 1 and thread_per_core >= 1
            result += f"{self.threads_per_rank} {thread_per_rank} "
            result += f"{self.threads_per_core} {thread_per_core} "
        return result
    
    def worker_str(self, workers):
        if not workers:
            return ""
        return f"-L {','.join(str(worker.id) for worker in workers)}"

class MPICHCommand(MPICommand):
    def __init__(self):
        # 64 independent jobs, 1 per core of a KNL node: -n64 -N64 -d1 -j1
        self.mpi = 'mpirun'
        self.nproc = '-n'
        self.ppn = '--ppn'
        self.env = '--env'
        self.cpu_binding = None
        self.threads_per_rank = None
        self.threads_per_core = None
    
    def worker_str(self, workers):
        return ""

class COOLEYMPICommand(MPICommand):
    def __init__(self):
        # 64 independent jobs, 1 per core of a KNL node: -n64 -N64 -d1 -j1
        self.mpi = 'mpirun'
        self.nproc = '-n'
        self.ppn = '--ppn'
        self.env = '--env'
        self.cpu_binding = None
        self.threads_per_rank = None
        self.threads_per_core = None
    
    def worker_str(self, workers):
        if not workers:
            return ""
        return f"--hosts {','.join(str(worker.id) for worker in workers)} "

supported_types = ['COOLEY', 'THETA']
if JobEnv.host_type in supported_types:
    MPIcmd = getattr(sys.modules[__name__], f"{JobEnv.host_type}MPICommand")
else:
    try:
        p = subprocess.Popen(['mpirun', '-npernode'], 
                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, _ = p.communicate()
    except:
        logger.warning("Warning: Balsam cannot popen mpirun..proceed at your own risk")
        MPIcmd = MPICHCommand
    else:
        if 'unrecognized argument npernode' in stdout.decode():
            logger.info("Assuming MPICH")
            MPIcmd = MPICHCommand
        else:
            logger.info("Assuming OpenMPI")
            MPIcmd = OPENMPICommand



# These must come before any other imports
# --------------
import argparse
import sys
from balsam.scripts.cli_commands import newapp,newjob,newdep,ls,modify,rm
from balsam.scripts.cli_commands import kill,mkchild,launcher,service,make_dummies
from balsam.scripts.cli_commands import init, which, server, submitlaunch, log
from balsam import __version__

def main():
    if not sys.version_info >= (3,6):
        sys.stderr.write("Balsam requires Python version >= 3.6\n")
        sys.exit(1)
    parser = make_parser()
    args = parser.parse_args()
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)
    args.func(args)

def config_launcher_subparser(subparser=None):
    if subparser is None:
        parser = argparse.ArgumentParser(description="Start Balsam Job Launcher.")
    else:
        parser = subparser

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--consume-all', action='store_true', help="Continuously run all jobs from DB")
    group.add_argument('--wf-filter', help="Continuously run jobs of specified workflow")
    parser.add_argument('--job-mode', choices=['mpi', 'serial'],
            required=True, default='mpi')
    parser.add_argument('--time-limit-minutes', type=float, default=0, 
                        help="Provide a walltime limit if not already imposed")
    parser.add_argument('--num-transition-threads', type=int, default=None)
    parser.add_argument('--gpus-per-node', type=int, default=None)
    return parser

def service_subparser(subparser=None):
    if subparser is None:
        parser = argparse.ArgumentParser(description="Start Balsam Job Launcher.")
    else:
        parser = subparser
    return parser

def make_parser():
    parser = argparse.ArgumentParser(prog='balsam', description="Balsam "+__version__)
    subparsers = parser.add_subparsers(title="Command line interface")


    # ADD APP
    # --------
    parser_app = subparsers.add_parser('app',
                                       help="add a new application definition",
                                       description="add a new application definition",
                                       )
    parser_app.set_defaults(func=newapp)
    parser_app.add_argument('--name', required=True)
    parser_app.add_argument('--executable', help='full path to executable', 
                            required=True)
    parser_app.add_argument('--preprocess', default='', 
                            help='preprocessing script with full path')
    parser_app.add_argument('--postprocess', default='',
                            help='postprocessing script with full path')
    parser_app.add_argument('--description', nargs='+')
    # -------------------------------------------------------------------


    # ADD JOB
    # -------
    parser_job = subparsers.add_parser('job',
                                       help="add a new Balsam job",
                                       description="add a new Balsam job",
                                       )
    parser_job.set_defaults(func=newjob)

    parser_job.add_argument('--name', required=True)
    parser_job.add_argument('--workflow', required=True,
                            help="A workflow name for grouping related jobs")
    parser_job.add_argument('--application', help='Name of the '
                            'application to use; must exist in '
                            'ApplicationDefinition DB', required=True)
    
    parser_job.add_argument('--wall-time-minutes', type=int, required=False, default=1)
    parser_job.add_argument('--num-nodes',
                            type=int, required=False,default=1,
                            help='Number of compute nodes on which to run MPI '
                            'job. (Total number of MPI ranks determined as '
                            'num_nodes * ranks_per_node).')
    parser_job.add_argument('--coschedule-num-nodes',
                            type=int, required=False,default=0)
    parser_job.add_argument('--node-packing-count',
                            type=int, required=False,default=1)
                            
    parser_job.add_argument('--ranks-per-node',
                            type=int, required=False,default=1,
                            help='Number of MPI ranks per compute node. '
                            '(Total MPI ranks calculated from num_nodes * '
                            'ranks_per_node. If only 1 total ranks, treated as serial '
                            'job).')
    
    parser_job.add_argument('--description', required=False, nargs='*', default=[])

    parser_job.add_argument('--threads-per-rank',type=int, default=1,
                            help="Equivalent to -d option in aprun")
    parser_job.add_argument('--threads-per-core',type=int, default=1,
                            help="Equivalent to -j option in aprun")

    parser_job.add_argument('--args', nargs='*', required=False, default=[],
                            help="Command-line args to the application")
    parser_job.add_argument('--post_handle_error', action='store_true',
                            help="Flag enables job runtime error handling by "
                            "postprocess script")
    parser_job.add_argument('--post_handle_timeout', action='store_true',
                            help="Flag enables job timeout handling by "
                            "postprocess script")
    parser_job.add_argument('--disable_auto_timeout_retry', action='store_true',
                            help="Flag disables automatic job retry if it has "
                            "timed out in a previous run")

    parser_job.add_argument('--input-files', nargs='*', required=False, 
                            default=['*'], help="Dataflow: filename patterns "
                            "that will be searched for in the parent job "
                            "working directories and retrieved for input. "
                            "[Ex: '*.log gopt.dat geom???.xyz' ]")

    parser_job.add_argument('--url-in', required=False,default='',
                            help='Input URL from which remote input files are copied.')
    parser_job.add_argument('--url-out',required=False,default='',
                            help='Output URL to which output files are copied.')
    parser_job.add_argument('--stage-out-files', nargs='*', required=False,default=[],
                            help="Filename patterns; matches will be "
                            "transferred to the destination specified " 
                            "by --url-out option")
    parser_job.add_argument('--env', action='append', required=False,
                            default=[], help="Environment variables specific " 
                            "to this job; specify multiple envs like " 
                            "'--env VAR1=VAL1 --env VAR2=VAL2'.  "
                            "Application-specific variables can instead be "
                            "given in the ApplicationDefinition to avoid "
                            "repetition here.")

    parser_job.add_argument('--yes', help='Skip prompt confirming job details.',
                            required=False,action='store_true')
    #--------------------------------------------------------------------


    # ADD DEP
    # -------
    parser_dep = subparsers.add_parser('dep',
                                       help="add a dependency between two existing jobs",
                                       description="add a dependency between two existing jobs"
                                       )
    parser_dep.set_defaults(func=newdep) 
    parser_dep.add_argument('parent', help="Parent must be finished before child")
    parser_dep.add_argument('child', help="The dependent job")
    #-------------------------------------------------------------------------


    # LS
    # ----
    parser_ls = subparsers.add_parser('ls', help="list jobs, applications, or jobs-by-workflow")
    parser_ls.set_defaults(func=ls)
    parser_ls.add_argument('objects', choices=['jobs', 'apps', 'wf', 'queues'], default='jobs',
                           nargs='?', help="list all jobs, all apps, or jobs by workflow")
    parser_ls.add_argument('--name', help="match any substring of job name")
    parser_ls.add_argument('--history', help="show state history / logs", action='store_true')
    parser_ls.add_argument('--id', help="match any substring of job id")
    parser_ls.add_argument('--state', help="list jobs matching a state")
    parser_ls.add_argument('--by-states', action='store_true', help="group job listing by states")
    parser_ls.add_argument('--wf', help="Filter jobs matching a workflow")
    parser_ls.add_argument('--verbose', help="Detailed BalsamJob info", action='store_true')
    parser_ls.add_argument('--tree', action='store_true', help="show DAG in tree format")
    # -----------------------------------------------------------


    # MODIFY
    # ------
    parser_modify = subparsers.add_parser('modify', help="alter job or application")
    parser_modify.set_defaults(func=modify)
    parser_modify.add_argument('id', help="substring of job or app ID to match")
    parser_modify.add_argument('attr', help="attribute of job or app to modify")
    parser_modify.add_argument('value', help="modify attr to this value")
    parser_modify.add_argument('--type', choices=['jobs', 'apps'], default='jobs', help="modify a job or AppDef")
    # -----------------------------------------------------------------------

    
    # RM
    # --
    parser_rm = subparsers.add_parser('rm', help="remove jobs or applications from the database")
    parser_rm.set_defaults(func=rm)
    parser_rm.add_argument('objects', choices=['jobs', 'apps'], help="permanently delete jobs or apps from DB")
    parser_rm.add_argument('--force', action='store_true', help="force delete")
    group = parser_rm.add_mutually_exclusive_group(required=True)
    group.add_argument('--name', help="match any substring of job name")
    group.add_argument('--id', help="match any substring of job id")
    group.add_argument('--all', action='store_true', help="delete all objects in the DB")
    # --------------------------------------------------------------------------------------------------


    # KILL
    # ----
    parser_kill = subparsers.add_parser('killjob', help="Kill a job without removing it from the DB")
    parser_kill.set_defaults(func=kill)
    parser_kill.add_argument('--id', required=True)
    parser_kill.add_argument('--recursive', action='store_true')
    # -----------------------------------------------------


    # MAKECHILD
    # ---------
    parser_mkchild = subparsers.add_parser('mkchild', help="Create a child job of a specified job")
    parser_mkchild.set_defaults(func=mkchild)
    parser_mkchild.add_argument('--name', required=True)
    parser_mkchild.add_argument('--workflow', required=True,
                            help="A workflow name for grouping related jobs")
    parser_mkchild.add_argument('--application', help='Name of the '
                            'application to use; must exist in '
                            'ApplicationDefinition DB', required=True)
    
    parser_mkchild.add_argument('--wall-minutes', type=int, required=True)
    parser_mkchild.add_argument('--num-nodes',
                            type=int, required=True)
    parser_mkchild.add_argument('--ranks-per-node',
                            type=int, required=True)
    
    parser_mkchild.add_argument('--description', required=False, nargs='*',
                            default=[])

    parser_mkchild.add_argument('--threads-per-rank',type=int, default=1,
                            help="Equivalent to -d option in aprun")
    parser_mkchild.add_argument('--threads-per-core',type=int, default=1,
                            help="Equivalent to -j option in aprun")

    parser_mkchild.add_argument('--args', nargs='*', required=False, default=[],
                            help="Command-line args to the application")
    parser_mkchild.add_argument('--preprocessor', required=False, default='',
                            help="Override application-defined preprocess")
    parser_mkchild.add_argument('--postprocessor', required=False, default='',
                            help="Override application-defined postprocess")
    parser_mkchild.add_argument('--post_handle_error', action='store_true',
                            help="Flag enables job runtime error handling by "
                            "postprocess script")
    parser_mkchild.add_argument('--post_handle_timeout', action='store_true',
                            help="Flag enables job timeout handling by "
                            "postprocess script")
    parser_mkchild.add_argument('--disable_auto_timeout_retry', action='store_true',
                            help="Flag disables automatic job retry if it has "
                            "timed out in a previous run")

    parser_mkchild.add_argument('--input-files', nargs='*', required=False, 
                            default=['*'], help="Dataflow: filename patterns "
                            "that will be searched for in the parent job "
                            "working directories and retrieved for input. "
                            "[Ex: '*.log gopt.dat geom???.xyz' ]")

    parser_mkchild.add_argument('--url-in', required=False,default='',
                            help='Input URL from which remote input files are copied.')
    parser_mkchild.add_argument('--url-out',required=False,default='',
                            help='Output URL to which output files are copied.')
    parser_mkchild.add_argument('--stage-out-files', nargs='*', required=False,default=[],
                            help="Filename patterns; matches will be "
                            "transferred to the destination specified " 
                            "by --url-out option")
    parser_mkchild.add_argument('--env', action='append', required=False,
                            default=[], help="Environment variables specific " 
                            "to this job; specify multiple envs like " 
                            "'--env VAR1=VAL1 --env VAR2=VAL2'.  "
                            "Application-specific variables can instead be "
                            "given in the ApplicationDefinition to avoid "
                            "repetition here.")

    parser_mkchild.add_argument('--yes', help='Skip prompt confirming job details.',
                            required=False,action='store_true')
    # -----------------------------------------------------


    # LAUNCHER
    # --------
    parser_launcher = subparsers.add_parser('launcher', help="Start a local instance of the balsam launcher")
    parser_launcher = config_launcher_subparser(parser_launcher)
    parser_launcher.set_defaults(func=launcher)
    # -----------------

    # SUBMIT-LAUNCH
    # -------------
    parser_submitlaunch = subparsers.add_parser('submit-launch', help="Submit a launcher job to the batch queue")
    parser_submitlaunch.add_argument('-n', '--nodes', type=int, required=True)
    parser_submitlaunch.add_argument('-t', '--time-minutes', type=int, required=True)
    parser_submitlaunch.add_argument('-q', '--queue', type=str, required=True)
    parser_submitlaunch.add_argument('-A', '--project', type=str, required=True)
    parser_submitlaunch.add_argument('--job-mode', type=str, choices=['serial', 'mpi'], required=True)
    parser_submitlaunch.add_argument('--wf-filter', type=str, default='')
    parser_submitlaunch.set_defaults(func=submitlaunch)
    
    # INIT
    # --------
    parser_init = subparsers.add_parser('init', help="Create new balsam DB")
    parser_init.add_argument('path', help="Path to Balsam DB directory")
    parser_init.set_defaults(func=init)
    # -----------------


    # SERVICE
    # -------
    parser_service = subparsers.add_parser('service', help="Start Balsam auto-scheduling service")
    parser_service.set_defaults(func=service)
    # -------------------------

    # DUMMIES
    # ---------
    parser_dummy = subparsers.add_parser('make_dummies')
    parser_dummy.add_argument('num', type=int)
    parser_dummy.set_defaults(func=make_dummies)
    
    # WHICH
    # ---------
    parser_which = subparsers.add_parser('which', help="Get info on current/available DBs")
    parser_which.add_argument('--list', action='store_true', help="list cached DB paths")
    parser_which.add_argument('--name', help='Look up DB path from a partial name')
    parser_which.set_defaults(func=which)
    
    # LOG
    # ---------
    parser_log = subparsers.add_parser('log', help="Quick view of Balsam log files")
    parser_log.set_defaults(func=log)
    
    # SERVER
    # ---------
    parser_server = subparsers.add_parser('server', help="Control Balsam server at BALSAM_DB_PATH")
    group = parser_server.add_mutually_exclusive_group(required=True)
    group.add_argument('--connect', action='store_true', help="connect to existing or start new server")
    group.add_argument('--reset', action='store_true', help="stop and start server")
    group.add_argument('--list-active-connections', action='store_true', help="see how many clients have active connection")
    group.add_argument('--list-users', action='store_true', help="list authorized Balsam users")
    group.add_argument('--add-user', type=str, help="add an authorized Balsam user")
    group.add_argument('--drop-user', type=str, help="drop a user from the DB")
    group.set_defaults(func=server)

    return parser


if __name__ == "__main__":
    main()



import sys
import time

if len(sys.argv) == 1:
    delay = 10
else:
    delay = int(sys.argv[1])

if 'parallel' in ' '.join(sys.argv):
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
else:
    rank = 0

print(f"Rank {rank} Sleeping for a long time...")
sys.stdout.flush()
time.sleep(delay)
if rank == 0: print("Done")



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import os
import random
import argparse
import time
import sys

parser = argparse.ArgumentParser()
parser.add_argument('--sleep', type=float, default=0)
parser.add_argument('--retcode', type=int, default=0)
args = parser.parse_args()

time.sleep(args.sleep)

num_sides = int(os.environ['BALSAM_FT_NUM_SIDES'])
num_ranks = int(os.environ['BALSAM_FT_NUM_RANKS'])

if num_ranks > 1:
    from mpi4py import MPI
    COMM = MPI.COMM_WORLD
    rank = COMM.Get_rank()
    print(f"Rank {rank}")
else:
    print("Rank 0")
    rank = 0

if rank == 0:
    for i in range(num_sides):
        side_length = random.uniform(0.5,5)
        with open(f"side{i}.dat", 'w') as fp:
            fp.write(str(side_length) + "\n")

sys.exit(args.retcode)



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import balsam.launcher.dag as dag
import sys

num_sides = sys.argv[1]
num_ranks = sys.argv[2]

current_job = dag.current_job
current_job.environ_vars=f"BALSAM_FT_NUM_SIDES={num_sides}:BALSAM_FT_NUM_RANKS={num_ranks}"
current_job.ranks_per_node = num_ranks
current_job.save(update_fields=['environ_vars', 'ranks_per_node'])

print("Hello from make_sides_pre")



import glob
import os
import sys
from pprint import pprint
import re
import subprocess
from balsam import setup, settings
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
from balsam.django_config.serverinfo import ServerInfo
from balsam.scripts import postgres_control
from django.db import connection

POSTGRES_REQUIRED = (9,6,4)
POSTGRES_MSG = \
f'''You need PostgreSQL {'.'.join(map(str, POSTGRES_REQUIRED))} or newer.
Ensure pg_ctl is in the search PATH, and double-check version with pg_ctl --version.
You can easily grab the Postgres binaries at:
https://www.enterprisedb.com/download-postgresql-binaries
'''

def check_postgres():
    p = subprocess.run(
        'pg_ctl --version',
        shell=True,
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=True
    )
    stdout = p.stdout.strip()
    pattern = re.compile(r'(\d+\.)?(\d+\.)?(\*|\d+)$')
    match = pattern.search(stdout).group()
    major, minor, *rest = match.split('.')
    if rest and int(rest[0])>0:
        version_info = (int(major), int(minor), int(rest[0]))
    else:
        version_info = (int(major), int(minor))
    if version_info >= POSTGRES_REQUIRED:
        print(f"PostgreSQL {version_info} meets required: {POSTGRES_REQUIRED}")
        return True
    else:
        print(f"PostgreSQL {version_info} does not meet required: {POSTGRES_REQUIRED}")
        return False

def postgres_init(serverInfo):
    if not os.path.exists(serverInfo.pg_db_path): # postgres data directory
        postgres_control.create_db(serverInfo)
    else:
        print("Postgres data directory already exists; will not create new DB")
        postgres_control.start_main(serverInfo.balsam_db_path)
        serverInfo.refresh()

def run_migrations():
    from django import db
    from django.core.management import call_command
    from balsam.django_config.db_index import refresh_db_index
    setup()
    print(f"DB settings:", settings.DATABASES['default'])
    call_command('makemigrations', interactive=True, verbosity=2)
    call_command('migrate', interactive=True, verbosity=2)
    refresh_db_index()
    try:
        from balsam.core.models import BalsamJob
        j = BalsamJob()
        j.save()
        j.delete()
    except:
        print("BalsamJob table not properly created")
        raise
    else:
        print("BalsamJob table created successfully")

if __name__ == "__main__":
    if not check_postgres():
        raise RuntimeError(POSTGRES_MSG)

    dbpath = sys.argv[1]
    os.environ['BALSAM_DB_PATH'] = dbpath
    serverInfo = ServerInfo(dbpath)

    postgres_init(serverInfo)
    run_migrations()
    connection.close()
    postgres_control.kill_server(serverInfo)

    base = dbpath[:-1] if dbpath[-1]=='/' else dbpath
    basename = os.path.basename(base)
    msg = f"  Successfully created Balsam DB at: {os.environ['BALSAM_DB_PATH']}  "
    print('\n' + '*'*len(msg))
    print(msg)
    print(f"  Use `source balsamactivate {basename}` to begin working.")
    print('*'*len(msg)+'\n')



from balsam.core import models
from django.conf import settings
from django.db.models import F, Sum, FloatField
from django.db.models.functions import Cast
import logging
logger = logging.getLogger(__name__)

JOB_PAD_MINUTES = 5
BalsamJob = models.BalsamJob
QueuedLaunch = models.QueuedLaunch

def create_qlaunch(queues):
    jobs = ready_query()
    qlaunch, to_launch = _pack_jobs(jobs, queues)
    if qlaunch and to_launch.exists():
        #qlaunch.wall_minuntes += JOB_PAD_MINUTES
        qlaunch.save()
        qlaunch.refresh_from_db()
        num = to_launch.update(queued_launch=qlaunch)
        logger.info(f'Scheduled {num} jobs in {qlaunch}')
        return qlaunch
    else:
        return None

    jobs = jobs.all()
    return qlaunch, jobs

def ready_query():
    return BalsamJob.objects.filter(
        lock='',
        queued_launch__isnull=True,
        state__in=['PREPROCESSED', 'RESTART_READY']
    )

def serial_node_count():
    jobs = ready_query()
    serial_jobs = jobs.filter(num_nodes=1, ranks_per_node=1)
    serial_jobs = serial_jobs.annotate(
        nodes=1.0/Cast('node_packing_count', FloatField())
        +Cast('coschedule_num_nodes', FloatField())
    )
    return serial_jobs.aggregate(total=Sum('nodes'))['total']

def box_pack(jobs, queues):
    import _packer
    Rect = _packer.Rect
    if serial_node_count() > SERIAL_PACK_THRESHOLD:
        jobs = ready_query.all()
    job_rects = [Rect(*job) for job in jobs.values_list(
                 ['num_nodes', ''])]
    for q in queues:
        pass

def dummy_pack(jobs, queues):
    '''Input: jobs (queryset of scheduleable jobs), queues(states of all queued
    jobs); Return: a qlaunch object (from which launcher qsub can be generated),
    and list/queryset of jobs scheduled for that launch'''
    if not queues: return None
    qname = list(queues.keys())[0]
    qlaunch = QueuedLaunch(queue=qname,
                           nodes=8,
                           job_mode='mpi',
                           prescheduled_only=False,
                           wall_minutes=60)
    return qlaunch, jobs.all()

_pack_jobs = dummy_pack



from balsam import settings
import subprocess
import logging
import glob
import os,sys,traceback
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

def validate_path(path):
    path = os.path.abspath(os.path.expanduser(path))
    if glob.glob(path):
        return path
    else:
        raise ValueError(f'Nothing matches {path} in filesystem')

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# - GridFTP implementation
GRIDFTP_PROTOCOL='gsiftp'
class GridFTPHandler:
   def pre_stage_hook(self):
      # check to see if proxy already exists
      p = subprocess.Popen([settings.GRIDFTP_PROXY_INFO,'-exists'])
      p.wait()
      if p.returncode is not 0: # valid proxy does not exist so create one
         command = str(settings.GRIDFTP_PROXY_INIT) + ' -verify -debug -bits 2048 -valid 96:00'
         logger.debug('command=' + command)
         try:
            p = subprocess.Popen(command.split(' '),stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            out,err = p.communicate()
         except OSError as e:
            logger.error('command failed with OSError, exception: ' + str(e))
            raise Exception('Error in pre_stage_hook, OSError raised')
         except ValueError as e:
            logger.error('command failed with ValueError, exception: ' + str(e))
            raise Exception('Error in pre_stage_hook, ValueError raised')
         except Exception as e:
            logger.error('command failed, exception traceback: \n' + traceback.format_exc() )
            raise Exception('Error in stage_in, unknown exception raised')

         if p.returncode:
            logger.error('command failed with return value: ' + str(p.returncode) + '\n  stdout: \n' + out + '\n stderr: \n' + err)
            raise Exception('Error in pre_stage_hook, return value = ' + str(p.returncode) )
         else:
            logger.debug('gridftp initialization completed, stdout:\n' + out + '\n stderr:\n' + err )

   def stage_in( self, source_url, destination_directory ):
      # ensure that source and destination each have a trailing '/'
      if source_url[-1] != '/':
         source_url += '/'
      if destination_directory[-1] != '/':
         destination_directory += '/'
      command = str(settings.GRIDFTP_GLOBUS_URL_COPY) + ' -dbg -nodcau -r %s %s' % (source_url, destination_directory)
      logger.debug('command=' + command )
      try:
         p = subprocess.Popen(command.split(' '),stdout=subprocess.PIPE,stderr=subprocess.PIPE)
         out,err = p.communicate()
      except OSError as e:
         logger.error('command failed with OSError, exception: ' + str(e))
         raise Exception('Error in stage_in, OSError raised')
      except ValueError as e:
         logger.error('command failed with ValueError, exception: ' + str(e))
         raise Exception('Error in stage_in, ValueError raised')
      except Exception as e:
         logger.error('command failed, exception traceback: \n' + traceback.format_exc() )
         raise Exception('Error in stage_in, unknown exception raised')
      if p.returncode:
         logger.error('command failed with return value: ' + str(p.returncode) + '\n  stdout: \n' + out + '\n stderr: \n' + err)
         raise Exception("Error in stage_in: %s" % str(p.returncode))

   def stage_out( self, source_directory, destination_url ):
      # ensure that source and destination each have a trailing '/'
      if source_directory[-1] != '/':
         source_directory += '/'
      if destination_url[-1] != '/':
         destination_url += '/'
      command = str(settings.GRIDFTP_GLOBUS_URL_COPY) + ' -dbg -nodcau -cd -r %s %s' % (source_directory, destination_url)
      logger.debug('command=' + command)
      try:
         p = subprocess.Popen(command.split(' '),stdout=subprocess.PIPE,stderr=subprocess.PIPE)
         out,err = p.communicate()
      except OSError as e:
         logger.error('command failed with OSError, exception: ' + str(e))
         raise Exception('Error in stage_out, OSError raised')
      except ValueError as e:
         logger.error('command failed with ValueError, exception: ' + str(e))
         raise Exception('Error in stage_out, ValueError raised')
      except Exception as e:
         logger.error('command failed, exception traceback: \n' + traceback.format_exc() )
         raise Exception('Error in stage_out, unknown exception raised')

      if p.returncode:
         logger.error('command failed with return value: ' + str(p.returncode) + '\n  stdout: \n' + out + '\n stderr: \n' + err)
         raise Exception("Error in stage_out: %s" % str(p.returncode))


# - Local implementation
LOCAL_PROTOCOL='local'
class LocalHandler:
   def pre_stage_hook(self):
      pass

   def stage_in(self, source_url, destination_directory):
      if source_url.strip().startswith('local:'):
          source = ''.join(source_url.split(':')[1:])
      else:
          source = source_url
      source = validate_path(source)
      dest = validate_path(destination_directory)
      assert os.path.isdir(dest), f'{dest} is not a valid destination directory'

      command = f'cp -p -r {source} {destination_directory}'
      logger.debug('transfer.stage_in: command=' + command )
      p = subprocess.Popen(command,stdout=subprocess.PIPE,stderr=subprocess.STDOUT, shell=True)
      stdout,stderr = p.communicate()
      if p.returncode != 0:
          raise Exception(f"Stage in process returned {p.returncode}: {stdout}")

   def stage_out(self, source_directory, destination_url):
      if destination_url.strip().startswith('local:'):
          dest = ''.join(destination_url.split(':')[1:])
      else:
          dest = destination_url
      source = validate_path(source_directory)
      dest = validate_path(dest)
      assert os.path.isdir(dest), f'{dest} is not a valid destination directory'

      command = f'cp -r {source} {dest}'
      logger.debug('transfer.stage_out: command=' + command )
      p = subprocess.Popen(command,stdout=subprocess.PIPE,stderr=subprocess.STDOUT, 
              shell=True)
      stdout,stderr = p.communicate()
      if p.returncode != 0:
         raise Exception("Error in stage_out: %d output:\n" % (p.returncode,stdout))

# - SCP implementation
SCP_PROTOCOL='scp'
class SCPHandler:
   def pre_stage_hook(self):
      pass

   def stage_in( self, source_url, destination_directory ):
      parts = urlparse.urlparse( source_url )
      command = 'scp -p -r %s:%s %s' % (source_url, destination_directory)
      logger.debug('transfer.stage_in: command=' + command )
      ret = os.system(command)
      if ret:
         raise Exception("Error in stage_in: %d" % ret)

   def stage_out( self, source_directory, destination_url ):
      # ensure that source and destination each have a trailing '/'
      command = 'scp -p -r %s %s' % (source_directory, destination_url)
      logger.debug('transfer.stage_out: command=' + command)
      ret = os.system(command)
      if ret:
         raise Exception("Error in stage_out: %d" % ret)


# - Generic interface

def get_handler(url):
    handlers = {
        GRIDFTP_PROTOCOL:GridFTPHandler,
        SCP_PROTOCOL    :SCPHandler,
        LOCAL_PROTOCOL  :LocalHandler
    }
    proto = url.split(':')[0]
    if proto in handlers.keys():
        handler_class = handlers[proto]
        handler = handler_class()
    else:
        url = validate_path(url)
        handler = LocalHandler()
    return handler

def stage_in(source_url, destination_directory):
    handler = get_handler(source_url)
    handler.pre_stage_hook()
    handler.stage_in(source_url, destination_directory)

def stage_out(source_directory, destination_url):
    handler = get_handler(destination_url)
    handler.pre_stage_hook()
    handler.stage_out(source_directory, destination_url)
